/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#ifndef GLDISP_H
#define GLDISP_H

/**
 * @file   gldisp.h
 * @author Eric C. Anderson
 * @date   Tue Aug 27 21:36:51 2002
 *
 * @brief  Tracking MCMC simulation using the GLUT interface.
 *
 * This is a header file for some routines that Eric C. Anderson
 * has written in order to (with luck) quickly and easily design
 * windowing systems using the GLUT interface to track MCMC simulations.
 *
 * The purpose of this package is to specify several data types and
 * to provide a few functions, so that a user can just define his
 * own rendering schemes for each window that he wants, and then the
 * functions herein will sort of take care of the rest.
 *
 * one of the main difficulties is in registering a callback to
 * render anew each active window (we want all open windows to be
 * updated on each sweep).
 *
 * \$Id: gldisp.h,v 1.4 2005/08/22 23:56:39 myrna Exp $
 *
 */
#ifndef GLUT_FOR_MCMC_HFLAG
#define GLUT_FOR_MCMC_HFLAG

/**
 * @defgroup GFMCMC GFMCMC
 *
 * GFMCMC (Glut for MCMC) is a library for visualizing and tracking MCMC
 * simulation.  It is implemented using glut (OpenGL Utilities Toolkit) which
 * is a window system independent programming interface to OpenGL.
 *
 * @ingroup  Libraries
 * @{
 */
#include "incs.h"

#include <GL/glut.h>

#include "gfmcmc_shortcuts.h"

/* {{{ macros */

/** the maximum number of windows that one can have open simultaneously */
#define MAX_WIN                    25
/** max length of a string for a color key label */
#define MAX_COLOR_KEY_LENGTH       150
/** max number of color schemes */
#define MAX_CS                     25
#define MAX_VIEWS                  50
#define MAX_VIEW_NAME_LENGTH       250
#define MAX_AXIS_LABEL_LENGTH      250
#define MAX_TITLE                  300
#define MAX_WIN_DEFS               300
#define MAX_SERIES_COLORS          40
#define MAX_INPUT_MSG_LENGTH       500
#define MAX_USER_DEFD_MENUS        30
#define LARGEST_WINDOW_ID          500

#define NUM_INTEGER_OPTIONS        500
#define NUM_FLOAT_OPTIONS          500

#define COLOR_PICKER_WINDOW_ID     100

#define ERROR_FILE                 "GFMCMC_Errors.txt"

/** @name Pixels surronding a window.
 * The numbers of pixels surrounding a window on the left, right, bottom and
 * top.  This is system dependent.  This is for the MacOS
 */
/*@{*/
#define GFM_L 6
#define GFM_R 6
#define GFM_B 6
#define GFM_T 23
/*@}*/

/** number of pixels in the top menu bar  */
#define GFM_TOP_BAR 21

/* }}} */
   
/* {{{ enums and structs */

enum axis_type
{
   X_AXIS,
   Y_AXIS
};

/** Constants for the position of axis. */
enum axis_position_t
{
   AXIS_LEFT_BELOW = 0,       /**< Put axis left and below. */
   AXIS_RIGHT_ABOVE           /**< Put axis right and above. */
};

/** Constants for the position of the label of axis. */
enum axis_nums_position_t
{
   AXIS_NUMS_LEFT_BELOW = 0,  /**< Left and below axis. */
   AXIS_NUMS_RIGHT_ABOVE      /**< Right and above axis. */
};

/** Constants for the position of the ticks of axis. */
enum axis_ticks_position_t
{
   AXIS_TICKS_LEFT_BELOW = 0, /**< Left and below. */
   AXIS_TICKS_RIGHT_ABOVE,    /**< Right and above. */
   AXIS_TICKS_CENTERED        /**< Centered. */
};

/** Constatns for some predefined color schemes. */
enum color_schemes_t
{
   FISHER_PRICE = 0,
   OL_GRAYBACK,
   DEEP_BLUE,
   GILBERT_AND_SHERMAN
};

/** Windows Type */
enum win_type
{
   STATIC_WIN,     /**<  a window that needn't change every cycle          */
   CHANGING_WIN    /**<  a window that should be updated after every sweep */
};

/** Legend Location */
enum LegendLocation
{
   LEGEND_LEFT,
   LEGEND_RIGHT,
   LEGEND_TOP,
   LEGEND_BOTTOM
};

/** Legend Type */
enum LegendType
{
   BARS,
   LINES
};


/*  typedefs and data structures */
/**
 * Color scheme structure.
 */
typedef struct
{
   /** Index of the color scheme */
   int ID;
   /** Number of colors available in the scheme */
   int Ncolor;
   /** Title of the color scheme */
   char Title[MAX_TITLE];
   /**
    * "Series" color values stored as three floats.  When plotting
    * things, the element of the series, subscripted by i gets the color
    * CS[i%Ncolor].
    */
   GLfloat **Series;
   /** Background color stored as three floats */
   GLfloat Background[3];
   /**
    * Color for axes, grids, ticks and axis labels, if any, stored as three
    * floats
    */
   GLfloat Axes[3];
   /** Color for text on the screen that is not axis labels */
   GLfloat Text[3];
   /** Color for a box that surrounds the legend */
   GLfloat LegendBoundary[3];
   /** Color for legend background */
   GLfloat LegendBackground[3];
} ColorScheme3f;

/**
 * Window Structure
 */
typedef struct
{
   /** index of this window definition */
   int WindowID;
   /** title of the window */
   char Title[MAX_TITLE];

   /** window type */
   enum win_type WT;

   /** denotes whether the window displays current values or averages, etc. */
   int QT;
   /** an array of names for colors (like "F1" or "Allele 5") */
   char **ColorKey;
   /**  the number of things that the different colors refer to */
   int *NumColorKeys;
   /**
    * Default color scheme id. This should be used to point at something in
    * the data structures of the MCMC program so that the default color
    * scheme may be applied to it in CreateWindow, this gets used in
    * changing the window settings.
    */
   enum color_schemes_t DefaultColorSchemeIdx;

   /**
    * Index of middle mouse menu.  If this is greater than 0, then we will
    * attach that menu to the right mouse button.
    */
   int MiddleMouseMenuIdx;

   /**
    * Attach main menu to the window.  This shall be a one if we want to NOT
    * attach the main menu to the window (this will then let use mouse
    * clicks on the PC.)
    */
   int NoMainMenu;

   /**
    * Pointer to the rendering function that should be called for the window.
    * The argument of DrawFunc is here specified to be a quant_type, the
    * top-level data structure must be available as a global variable to pass
    * in to the function
    */
   void (*DrawFunc)(void);

   /**
    * Function to be called upon window creation
    */
   void (*AtCreation)(void);

   /**
    * Function to be called upon window destruction
    */
   void (*AtDestruction)(void);

   void (*ExtraFunction)(void);
   void (*KeyboardFunc)(unsigned char, int, int);
} window_struct;

/**
 * Legend setting.
 */
typedef struct
{
   int DrawLegend;           /**< 1 if a legend should be drawn, 0
                               otherwise */
   int AlreadyResized;       /**<  0 for no and 1 for yes */

   /** Legend Type */
   enum LegendType Type;
   /** Legend location */
   enum LegendLocation Location;

   /**
    * The percentage of the height of the window that the legend should take
    * up.
    */
   GLfloat PercentH;
   /**
    * The percentage of the width of the window that the legend should take
    * up.  Notice that if we have a function that can change this during
    * program execution, we must be careful to ensure that it is always set
    * to the proportion of the @em original screen.
    */
   GLfloat PercentW;

   /**
    * The legend font size.
    *
    *   -  0 = GLUT_BITMAP_HELVETICA_10
    *
    *   -  1 = GLUT_BITMAP_HELVETICA_12
    *
    *   -  2 = GLUT_BITMAP_HELVETICA_18
    */
   int FontSize;

   /**
    * Separation between bars as a proportion of their width.
    */
   GLfloat xs;
   /**
    * Separation between bars as a proportion of their height. (??)
    * Note that in the case of line legends, these will have some meaning,
    * too.
    */
   GLfloat ys;

   /*@{*/
   /**
    * Margins around the area in which the legend is rendered in the height or
    * width.
    */
   GLfloat BufferH;
   GLfloat BufferW;
   /*@}*/

} LegendSettings;

/**
 * Structure to define a @em clip, a 3-D cube.
 */
typedef struct
{
   GLfloat xlo;
   GLfloat xhi;
   GLfloat ylo;
   GLfloat yhi;
   GLfloat zlo;
   GLfloat zhi;
} clipping_struct;

/**
 * Fitting Structure.  This is used to automatically expand or contract the
 * clipping volume so that the contents of window (primarily I am thinking
 * here about plotted points or curves) will fit into it.  It has several
 * flags in it, to keep it from trying to change the clipping volume if it
 * does not know how big to make it.  It is the responsibility of the user to
 * determine the max and the min of plotted points in the function that is
 * being called to plot those points.
 */
typedef struct
{
   int AlwaysFit;             /**< if 1 it will resize the thing on every
                                iteration (seldom used) */
   int FitToExtremaNow;       /**< if it is a one, it will try to re-clip,
                                but it also requires that the new mins and
                                maxes have been set */
   int FitToDefaultsNow;      /**< this being 1 will cause the clipping
                                volume to be set back to the default
                                clipping */
   int ExtremaSet;            /**< this should get to set to 1 after the max
                                ranges have been set and then should be set
                                back to 0 after the clipping volume has been
                                changed. */

   /**
    * The "extra bits" between the extreme points and the edge of the
    * clipping volume.  So, these values are not really clipping edges
    * --they are proportions by which the clipping volume should exceed the
    * extreme in all directions.
    */
   clipping_struct *Padding;

   /**
    * The max and min in every direction upon initialization of a window.
    * These are set to the initial window clips.
    */
   clipping_struct *Extrema;

   /**
    * The default values of the clipping struct. When a window gets clipped
    * to these values, it does not include the Padding.
    */
   clipping_struct *Default;

} fitting_struct;

/**
 * The Axis Structure.
 */
typedef struct
{
   int DrawIt;                /**< 1 if it should be drawn, 0 otherwise */
   int ForceThroughPoint;     /**< 1 if the axis should be forced through a
                                point */

   GLfloat Intersect;         /**< The point to force it through. This
                                controls the Y position of the X axis and
                                the X position of the Y axis */

   /**
    * Which edge the axis should be place off of if this it is not forced
    * through a point.  It takes one of AXIS_LEFT_BELOW or AXIS_RIGHT_ABOVE
    */
   int AxisEdge;

   /**
    * The position of the axis. This gets set by DrawXaxis() or DrawYAxis()
    * and tells where it is in the scene.  Then the ticks and grids can know
    * where it is too.
    */
   GLfloat AxisPos;

   /**
    * the proportion of the screen from the left side that the Y axis should
    * be put and from the bottom that the X axis should be put.
    */
   GLfloat PropFromEdge;

   /**
    * 1 if there is an upper limit beyond which no axis gets drawn  otherwise
    * the axis gets drawn across the entire clipping region
    */
   int ImposeUpperLimit;

   /** The upper limit if there is @ref ImposeUpperLimit is true. */
   GLfloat UpperLimit;

   /**
    * 1 if there is a lower limit beyond which no axis gets drawn  otherwise
    * the axis gets drawn across the entire clipping region
    */
   int ImposeLowerLimit;

   /** The lower limit if there is @ref ImposeLowerLimit is true. */
   GLfloat LowerLimit;

   /**
    * Determines where relative to the axis the ticks get drawn.
    * Uses one of: AXIS_TICKS_LEFT_BELOW or AXIS_TICKS_RIGHT_ABOVE or
    * AXIS_TICKS_CENTERED
    */
   enum axis_ticks_position_t TickPos;

   /** 1 if the label should be drawn */
   int WriteLabel;

   /**  for the name that will appear below the axis */
   char Label[MAX_AXIS_LABEL_LENGTH];

   int OptimalNumMajor;       /**< the user can set this to determine the
                               * number of Major ticks and hence number
                               * labels along the axis */
   GLfloat MajorProp;         /**< proportion of the viewable screen area
                               * that the major ticks should be */
   GLfloat MinorProp;         /**< proportion of the Major tick size that the
                               * minor ticks should be */


   /*
    * the function "gfmDetermineMajorTicksX (or Y)" sets the following values
    */

   GLfloat MajorSep;         /**< the distance between major ticks */
   GLfloat MajorLo;          /**< the lowest value for which a tick/number
                               label will be drawn */
   GLfloat MajorHi;          /**<
                              * the highest value for which a tick/number
                              * label will be drawn */
   GLfloat MinorSep;         /**<
                              * the distance between the minor ticks.  They
                              * will get drawn in between MajorLo and MajorHi
                              */

   int SigFigs;              /**< the number of digits right of the decimal
                               place that should be printed */

   int DrawGrids;            /**< 1 if stippled grids should be drawn at
                               Major Tick spacing */

   int DrawNumLabels;        /**< 1 if the number labels should be drawn */
   GLfloat NumbersDist;      /**< distance from axis for the number labels.
                              * Down (or to the left) if positive and above
                              * (to the right) if negative, as a proportion
                              * of the viewable width or height */
   GLfloat NumsHeight;       /**< height of axis number label text as
                              * proportion of total height */
} XY_axis_struct;


/**
 * Structure to define variables dealing with drawing a rectangle on the
 * screen that can be zoomed into.
 */
typedef struct
{
   int DrawDZBox;            /**< 1 if we should draw a DragZoomBox */

   GLfloat downX;            /**< coordinates of the mouse upon GLUT_DOWN
                              * mouseclick */
   GLfloat downY;            /**< coordinates of the mouse upon GLUT_DOWN
                              * mouseclick */

   GLfloat upX;              /**< coordinates of the mouse upon GLUT_UP
                              * mouseclick @em AND while dragging the mouse.
                              */
   GLfloat upY;              /**< coordinates of the mouse upon GLUT_UP
                              * mouseclick @em AND while dragging the mouse.
                              */
   /**
    * @todo eventually will want to add a stippling pattern thing here
    */
} drag_zoom_struct;

/**
 * Windows setting structure.
 */
typedef struct
{
   /**
    * here is an entry that does NOT get written out to file when recording
    * window settings!
    */
   int MenuEntry;

   /** variables to store the x, y, and z coordinate limits */
   clipping_struct *Clips;

   ColorScheme3f *ColorScheme;     /**< an array of the colors used by the
                                     window */

   /** The legend */
   LegendSettings *Legend;

   /**
    * variables to help with resizing the window to fit available points on
    * screen
    */
   fitting_struct *Fits;

   /* variable to hold info about the axes. */
   /** The X axis structure */
   XY_axis_struct *Xaxis;
   /** The Y axis structure */
   XY_axis_struct *Yaxis;

   /** The zoom structure */
   drag_zoom_struct *DZ;

   /**
    * X coordinate of the mouse for recording the position of the mouse during
    * MotionFunc() callbacks.
    */
   GLfloat XCoord;
   /**
    * Y coordinate of the mouse for recording the position of the mouse during
    * MotionFunc() callbacks.
    */
   GLfloat YCoord;

   /** The X position of the window (to the left upper corner), in pixels */
   int XPos;
   /** The Y position of the window (to the left upper corner), in pixels */
   int YPos;
   /** The width of the window, in pixels */
   int Width;
   /** The height of the window, in pixels */
   int Height;
   
   /*  the ''bounce-back" versions of these things.  They get set */
   /*  to the original window values when it is created, but */
   /*  then they will get reset to current values when the */
   /*  backslash "\" command is issued in the window, and it gets */
   /*  resized, to full screen.  The next "\" command to the window */
   /*  returns it from whence it came before the auto-full-size resizing. */
   int bbXPos;  
   int bbYPos; 
   int bbWidth;    
   int bbHeight;  
   
   /** starts as zero, but gets reset when it gets full-sized.s */
   int IsFullSize; 
   /** the number of columns in Bar-Type graphs */
   int NumCols;
   /** The Element currently selected by a mouse click */
   int SelectedElement;  
   /**
    * For screens where you view just one of many possible things
    * in the window, this tells which element of the array to view. 
    */
   int DisplayedElement; 

   /**  for user-defined options */
   int *IntegerOptions;  
   float *FloatOptions;
} window_settings;

/**
 * Viewing structure
 */
typedef struct
{
   /** the number of open windows in the view */
   int NumOpen;    
   /** the name of this particular view */
   char Name[MAX_VIEW_NAME_LENGTH];        

   /** The width in pixels of the original window the view was made on. These
    * will allow scaling to smaller or larger screens.
    */
   int OrigWidth;
   /** The height in pixels of the original window the view was made on. These
    * will allow scaling to smaller or larger screens.
    */
   int OrigHeight;

   /** the ID's of the different windows that are open */
   int *IDs;       
   /** the window settings of the open windows */
   window_settings **WS;  
} view_struct;

/* }}} */

/* {{{ global variables */

/*  some global variables that have to do with the screen size and the size */
/*  of borders around windows, and the height of title bars for them, etc. */
/*  Screen Height.  This is the number of pixels of screen height */

/** Screen height in pixels. */
GLOBAL int gSH;  

/** Header height in pixels.  This is the height taken up by the menu bar at
 * the top of screen.  On Mac OS 9, it seems to be about 20. */
GLOBAL int gHH;

/** Screen width */
GLOBAL int gSW;  
/** Height of a window top border in pixels (about 20 on the mac) */ 
GLOBAL int gWTB;
/** Size of the other window borders in pixels. (about 5 or 6 on the mac) */
GLOBAL int gWB;  

/**
 * Command console window width---the window width of the window that lists
 * all the views, etc.
 */
GLOBAL int gCCWW;  

/*  GLOBAL Variables of the Win Def Set and Views sort: */

GLOBAL int gWinIDReverse[LARGEST_WINDOW_ID + 1];
GLOBAL int gNumWinDefs;
GLOBAL window_struct *gWinDefs[MAX_WIN_DEFS + 1];  /*  the array of windows_structs that hold the  */
/*  definitions of the windows.  Will be allocated to dynamically */
/*  during run time when it is known how many different types of  */
/*  windows one can display */
GLOBAL window_settings *gInitWinSettings[MAX_WIN_DEFS + 1];
GLOBAL window_struct *gWindowsArray[MAX_WIN + 1];
GLOBAL window_settings *gWindowsSettings[MAX_WIN + 1];
GLOBAL int gNumViews;                     /*  the number of predefined views */
GLOBAL view_struct *gViews[MAX_VIEWS + 1];  /*  for PREDEFINED Views */
GLOBAL int gNumUDViews;           /*  for user-defined views */
GLOBAL view_struct *gUDViews[MAX_VIEWS + 1];
GLOBAL int gNumOpenWindows;


GLOBAL int gGo;  /*  global variable for Stopping and Going with the space bar toggle */
GLOBAL int gDoASingleSweep;  /*  for controlling "Single Sweep Mode" */
GLOBAL int gNumSweeps;
GLOBAL int gNumSweepsAfterBurnIn;

/*  silly flag for whether or not we are currently processing the console */
/*  window in UserDefd_DefineWindows() */
GLOBAL int gCurrentlyProcessingConsole;

/*  global variables for storing menu numbers */
GLOBAL int gMainMenu;
GLOBAL int gChainControlSubMenu;
GLOBAL int gSwitchWindowSubMenu;
GLOBAL int gCreateNewWindowsSubMenu;
GLOBAL int gViewsSubMenu;
GLOBAL int gCloseWindowsSubMenu;
GLOBAL int gPredefinedViewsSubMenu;  /*  this holds all the views that the developer has defined. */
GLOBAL int gUserDefdViewsSubMenu; /*  for views that the user defines */



GLOBAL int gLegendControlSubMenu;
GLOBAL int gUserDefdColorSubMenu;
GLOBAL int gPreDefColorSubMenu;
GLOBAL int gAutoFitSubMenu;
GLOBAL int gBringToFrontSubMenu;
GLOBAL int gQuitSubMenu;
GLOBAL int gColorSchemeMenu;
GLOBAL int gUserDefdMenus[MAX_USER_DEFD_MENUS];



/*  global arrays of pointers to color schemes that are available */
GLOBAL ColorScheme3f *gPreDefdColorSchemes[MAX_CS + 1];
GLOBAL ColorScheme3f *gUserDefdColorSchemes[MAX_CS + 1];
GLOBAL ColorScheme3f *gColorSchemes[MAX_CS + 1];
/*  global variables for keeping track of color schemes in the above arrays */
GLOBAL int gNumPreDefdColorSchemes;
GLOBAL int gNumUserDefdColorSchemes;


/*  global variables for defining and setting the color picker window */
/*  and for identifying it */
GLOBAL window_struct *gColorPickerWinDef;
GLOBAL window_settings *gColorPickerWinSettings;
GLOBAL int gColorPickerWindowNumber;


/*  global variables to have while running the "Color Picker" window */
GLOBAL int gCSPState;   /*  Color scheme picker state */
GLOBAL int gCSPHiState;  /*  the highest value that CSPState has yet reached */
GLOBAL int gCSPExit;
GLOBAL int gCSPComplete;  /*  gets a one if the whole color scheme has been done right */
GLOBAL int gCSPHiN;
GLOBAL int gCSPCurrentN;
GLOBAL int gCSPSpaceForN;  /*  records how much space has been allocated for the Series */
GLOBAL GLfloat gCSPr;  /*  red, blue and green values to make the instructions shimmer */
GLOBAL GLfloat gCSPb;  /*  so that they can be seen on any background color        */
GLOBAL GLfloat gCSPg;      /*  these are actually arguments to a sin function */
/*  pointer to the array of 216 colors. */
GLOBAL GLfloat **g216Colors3f;


/*  global variables for keeping track of the creation of a string on screen */
GLOBAL int gCurrentCharacter;
GLOBAL char *gInputString;  /*  this should get pointed to wherever the
                                string's destination should be. */
GLOBAL char gMsg[3][MAX_INPUT_MSG_LENGTH];
GLOBAL void (*gUponStringCompletion)(void);  /*  global pointer to function
                        that should be called when string input has concluded */
GLOBAL int gWindowRequestingStringInput;  
                                    /*  gUponStringCompletionMayWishToUseThis */
GLOBAL window_struct *gStringInputWindowDef;
GLOBAL window_settings* gStringInputWindowSettings;

/**
 * Global strings that get defined by appending the appropriate ending (i.e.
 * PreDefdViews.txt) to the prefix name provided by the developer.
 */
GLOBAL char gPredDefdViewsFileName[100];
GLOBAL char gUserDefdViewsFileName[100];
GLOBAL char gUserDefdColorSchemesFileName[100];


/**
 * This variable tell us whether we should use the current random number seeds
 * when we initialize the chain or not
 */
GLOBAL int gUseSameSeeds;

/* }}} */

/** @name User functions.
 * Prototypes for functions to be defined by the user.
 */
/*@{{{1*/

GLOBAL void gfmUserDefd_ResetAllAverages (void);

GLOBAL void gfmUserDefd_InitializeChain(void);

GLOBAL void gfmUserDefd_OneStepByGlobals(void);

GLOBAL void gfmUserDefd_DefineWindows(void);

GLOBAL void gfmUserDefd_DefineMenus(void);  /*  to define menus for the right
                                                click of the mouse */

GLOBAL void gfmUserDefd_LastWords(void);

/*@}*/

/* {{{ Functions defined in the GFMCMC library. */

/* {{{ initialization and memory allocation */

GLOBAL void gfmInitGFM(void);
GLOBAL void gfmUpdateImages(void);
GLOBAL void gfmSilly(void);
GLOBAL void gfmIdleFunc(void);

GLOBAL window_struct *gfmAllocWindowStruct(void);
GLOBAL window_settings *gfmAllocWindowSettings(void);
GLOBAL void gfmFreeWindowSettings(window_settings *W);
GLOBAL void gfmCopyWindowSettings(window_settings *From, window_settings *To);
GLOBAL view_struct *gfmAllocView(int nwin);
GLOBAL void gfmResizeToFull(void);

GLOBAL void gfmInitDragZoom(drag_zoom_struct *Z, int DrawBox, 
                            GLfloat x, GLfloat y);

/* }}} */

/** @name Windows functions.
 * Functions to create, destroy and move windows.
 */
/*@{{{1*/

/**
 * Destorys a window.
 *
 * This function destroys the window identified by @a CW, also erases its
 * entry in the "bring to front" menu, and decrements the MenuEntry field of
 * all the windows' above it (that is, all the windows created after @a CW).
 * This function also frees the memory pointed by @c gWindowsSettings[CW].
 *
 * @param CW the identifier of the window to be destroyed.
 */
GLOBAL void gfmDestroyWindow(int CW);

/**
 * Creates a window.
 * This function creates a window with the specifications in @a W having
 * settings initialized as in @a InitWS, unless InitWS == NULL, in which case
 * the default values set in ::gfmAllocWindowSettings are used.
 *
 * @param W window specification.
 * @param InitWS initial window setting.
 */
void gfmCreateWindow( window_struct *W, window_settings *InitWS );

/**
 * Create a window by index.
 * This function is used to be a menu callback, so that the menu options can
 * open windows.
 *
 * @param idx the index of the window to be opened.
 */
void gfmCreateWindowByIndex( int idx );

/**
 * The basic function to open the windows associated with the L-th view
 * in the array of views.
 * @param L the index of the desired view.
 * @param V an array of views.
 */
void gfmCreateWindowsForView(int L, view_struct **V);

/**
 * The menu callback for predefined menus.
 * @param L the index of the desired view.
 */
void gfmCreateWindowsForPreDefdViews(int L);

/**
 * The menu callback for user-defined menus.
 * @param L the index of the desired view.
 */
void gfmCreateWindowsForUserDefdViews(int L);

/**
 * Bring the window to front.
 * @param W the identifier of the window to be brought to front.
 */
void gfmBringWindowToFront(int W);


/*@}*/

/* {{{ Recording New Views, Reading and Writing Settings and Views to Files */

GLOBAL void gfmRecordNewView(int Dummy);
GLOBAL void gfmUponEntryOfViewName(void);
GLOBAL void gfmWriteWindowSettings(window_settings *WS, FILE *F);
GLOBAL void gfmReadWindowSettings(window_settings *WS, FILE *F);
GLOBAL void gfmReadWindowSettingsFromFile(window_settings *WS, FILE *F);
GLOBAL void gfmWriteViewsToFile(view_struct **V, char *FileName);
GLOBAL void gfmReadViewsFromFile(view_struct **V, char *FileName, int *Num,
                                 int Menu);
GLOBAL void gfmScaleTheView(GLfloat w, GLfloat h);
GLOBAL void gfmScaleWindowPosAndSizeViewWise(window_settings *W,
                                             GLfloat w, GLfloat h);
GLOBAL void gfmWriteFittingStructToFile(fitting_struct *F, FILE *out);
GLOBAL void gfmReadFittingStructFromFile(fitting_struct *F, FILE *in);
GLOBAL void gfmWriteClippingStructToFile(clipping_struct *C,FILE *out);
GLOBAL void gfmReadClippingStructFromFile(clipping_struct *C,FILE *in);
GLOBAL void gfmWriteXYAxisSettingsToFile(XY_axis_struct *A, FILE *out);
GLOBAL void gfmReadXYAxisSettingsToFile(XY_axis_struct *A, FILE *out);

/* }}} */

/* {{{ viewable area */

/**
 * Get the aspect ratio of things on the screen for 2-D images.  This is
 * Width Over Height of the current window.
 *
 * @return the aspect ratio (width/height) of current window.
 */
GLOBAL GLfloat gfmWoH(void);

GLOBAL void gfmUnitSquareWithBorder(window_settings *W, float border);

GLOBAL void gfmCopyClippings(clipping_struct *From,clipping_struct *To);

GLOBAL void gfmSetPaddedExtremeClips(window_settings *WS);

GLOBAL void gfmChangeSize(GLsizei w, GLsizei h);

GLOBAL void gfmClipOrtho(void);

GLOBAL void gfmChangeClippedByProportion(GLfloat P, clipping_struct *Clips,
                                         int Direction);

GLOBAL void gfmTranslateClippedByProportion(GLfloat P, clipping_struct *Clips,
                                            int Direction);

GLOBAL void gfmZoomToDragBox(void);

/* }}} */

/* {{{ weird things */

GLOBAL void gfmResizeAllForProperInit(void);
GLOBAL void gfmResizeForProperInit(void);
GLOBAL void gfmColor3fvAlphaf(GLfloat *c, GLfloat a);

/* }}} */

/* {{{ functions for processing keyboard input */

GLOBAL void gfmProcessNormalKeys(unsigned char key, int x, int y);
GLOBAL void gfmProcessSpecialKeys(int key, int x, int y);

/* }}} */

/* {{{ functions for determining the x, y coordinates of mouse clicks */

GLOBAL GLfloat gfmReturnXCoord(int XPix);
GLOBAL GLfloat gfmReturnYCoord(int YPix);
GLOBAL void gfmMouseLocation(int x, int y);
GLOBAL void gfmStandardMouseClick(int button, int state, int x, int y);

/* }}} */

/* {{{ defining the menus */

GLOBAL void gfmDefineGFMMenus(void);
GLOBAL void gfmKeysMirroredInMenuFunc(int Num);

/* }}} */

/* {{{ color scheme memory allocation */

GLOBAL ColorScheme3f *gfmAllocColorScheme(int ncolor);
GLOBAL void gfmFreeColorScheme(ColorScheme3f *C);

/* }}} */

/* {{{ color picker window functions */

GLOBAL void gfmDefineColorPickerWindow(void);
GLOBAL void Define216Colors(void);
GLOBAL void gfmDrawColorPicker(void);
GLOBAL void gfmColorPickerKeyboardFunc(unsigned char key, int x, int y);
GLOBAL void gfmInitiateColorPicking(int input);
GLOBAL void gfmCallAfterColorSchemeStringInput(void);

/* }}} */

/* {{{ functions for setting color schemes in windows */

GLOBAL void gfmSetToUserDefdColorScheme(int input);
GLOBAL void gfmSetToPreDefdColorScheme(int state);

/* }}} */

/* {{{ functions for writing and reading color schemes to file */

/**
 * Write a color scheme to a file stream.
 * @param C the color scheme to write.
 * @param F the file pointer to write to.
 */
GLOBAL void gfmWriteColorScheme3f( ColorScheme3f *C, FILE *F );

/**
 * Read a color scheme from a file stream.
 * @param C the color scheme read will be save in @a C.
 * @param F the file stream to read from.
 */
GLOBAL void gfmReadColorScheme3f( ColorScheme3f *C, FILE *F );
GLOBAL void gfmWriteColorSchemesToFile(ColorScheme3f **C, int ncolor,
                                       char *FileName );
GLOBAL void gfmWriteColorSchemesToFileAsSource(ColorScheme3f **C, int,
                                               char *FileName );
GLOBAL void gfmReadColorSchemesFromFile(ColorScheme3f *Temp[MAX_CS + 1],
                                     const char *FileName, int *NumC, int Menu);
GLOBAL void gfmWriteColorScheme3fAsSource(ColorScheme3f *C, FILE *F);

/* }}} */

/* {{{ legend functions */

/**
 * Copy legend setting.
 * @param From the LegendSettings struct to copy from.
 * @param To the LegendSettings struct to copy to.
 */
void gfmCopyLegendSettings( LegendSettings *From, LegendSettings *To );

/**
 * Draw a legend on the screen.
 * The position (at the top, bottom, left or right of the screen) depends on
 * the values in LegendSettings.  This does this by enlarging the @e margin
 * at the top or the bottom of the screen so that it fills "Percent" (part of
 * the LegendSettings struct) of the height of the screen, and the rest of the
 * stuff gets scrunched up a little bit.  If there is stuff that is outside of
 * the current clipping volume, then it will get drawn over.
 *
 * @a L->AlreadyResized should be zero on the first time this function
 * gets called in a rendering function for any of the screens.
 *
 * @param ncolor the number of different color key tables.
 * @param Names
 * @param Clips
 * @param C the color scheme.
 * @param L the legend setting.
 */
void gfmDrawLegend( int ncolor, char **Names, clipping_struct *Clips,
                    ColorScheme3f *C, LegendSettings *L);

void gfmZapTheLegend(window_settings *WS);

/* }}} */

/* {{{ functions for rendering text on the screen */

/**
 * Renders a string in the proportional roman font.
 * @param s the string to be rendered.
 * @param h the height (top to bottom) of the max-height xter (in the
 *          alphabet) in real units.
 * @param x x coordinate in the plane.
 * @param y y coordinate in the plane where the bottom-line of the text
 *          should be.
 * @param On how the text is to be aligned.
 * <dl>
 *  <dt>0</dt><dd>left edge of first letter</dd>
 *  <dt>1</dt><dd>center of first letter</dd>
 *  <dt>2</dt><dd>center of the string</dd>
 *  <dt>3</dt><dd>center of last letter</dd>
 *  <dt>4</dt><dd>right edge of last letter</dd>
 * </dl>
 * @param r the rotation angle of the text. Zero is horizontal to the right.
 * @param WoH "width over height" and aspect ratio of sorts.
 */

GLOBAL void gfmStrokeString( char *s, GLfloat h, GLfloat x, GLfloat y,
                             int On, GLfloat r, GLfloat WoH );

GLOBAL void gfmRenderBitmapString(float x, float y, void *font, char *string);
GLOBAL void gfmRenderBitmapCharacters(float x, float y, void *font,
                                      char *string, int Num);
/* }}} */

/* {{{ collecting string input from the user through a window */

GLOBAL void gfmAtCreationOfStringInput(void);
GLOBAL void gfmStringInputKeyboardFunc(unsigned char key, int x, int y);
GLOBAL void gfmHandleStringInput(char *M1,char *M2, char *M3, char *String,
                                 void (*FinishUpFunc)(void) );
GLOBAL void gfmDrawStringInputWindow(void);
GLOBAL void gfmDefineStringInputWindow(void);

/* }}} */

/* {{{ functions for positioning, drawing, and selecting bar graphs */

GLOBAL GLfloat gfmBarGraphYCoord(int n, GLfloat h, GLfloat ys, int i);
GLOBAL GLfloat gfmBarGraphXCoord(int n, GLfloat w, GLfloat xs, int i);

GLOBAL int gfmBarGraphIdxFromXY(int n, int ntotal, GLfloat w, GLfloat h,
                          GLfloat xs, GLfloat ys, double xcoord, double ycoord);

GLOBAL void gfmSetParsForBarGraphs(int ntotal, int C, GLfloat xs, GLfloat ys,
                          int *n, GLfloat *w, GLfloat *h);

/**
 * Draws a leaf on a rectangle that represents the proportions of a total.
 *
 * @param x  the x coordinate of the upper left of the rectangle.
 * @param y  the y coordinate of the upper left of the rectangle.
 * @param w  the width of the rectangle.
 * @param h  the height of the rectangle.
 * @param value the value currently being represented.
 * @param total the value represented by the sum of all the values in the
 * array.
 * @param Color what color to use
 */
GLOBAL void gfmDrawRectLeaf6f(GLfloat x, GLfloat y, GLfloat w, GLfloat h,
                              GLfloat value, GLfloat total, GLfloat *Color );

GLOBAL void 
gfmAppendToFile( char *FileName, char *String );

/* }}} */

/* {{{ axes */

GLOBAL void gfmDrawXaxis(void);
GLOBAL void gfmDrawYaxis(void);
GLOBAL void gfmCopyXYAxisStructs(XY_axis_struct *From, XY_axis_struct *To);
GLOBAL void gfmDetermineMajorTicksX(void);
GLOBAL void gfmDrawTicksX(void);
GLOBAL void gfmDetermineMajorTicksY(void);
GLOBAL void gfmDrawTicksY(void);
GLOBAL void gfmDrawXAxisNumLabels(void);
GLOBAL void gfmDrawYAxisNumLabels(void);
GLOBAL void gfmDrawXYAxes(void);

/* }}} */

void gs_handle_extrema( window_settings **ws, GLfloat xl, GLfloat xh,
                                              GLfloat yl, GLfloat yh );

void gfm_setup_window( int id, const char *title );

/**
 * Set the prefix of various file names.
 *
 * This function should be called in UserDefd_DefineWindows to name
 * the predefined view files and other output files with a prefix telling
 * us which application they came from.
 *
 * @param prefix The prefix name (less than 20 characters).
 */
void gfm_set_output_file_prefix( const char *prefix );

/**
 * Get the initial setting of current window.
 *
 * @return a pointer to the initial setting struct.
 */
window_settings * gfm_get_current_init_window_setting( );

/**
 * Get the current window.
 *
 * @return a pointer to current window struct.
 */
window_struct * gfm_get_current_window();

/**
 * Set the drawing function for current window.
 *
 * @param myfunc a user-defined drawing function.
 */
void gfm_set_draw_func(void (*myfunc)(void));

/**
 * Set the extra function for current window.
 *
 * @param myfunc a user-defined funtion.
 */
void gfm_set_extra_func(void (*myfunc)(void));

void gfm_set_keyboard_func(void (*myfunc)(unsigned char, int, int));

XY_axis_struct * gfm_get_current_axis( enum axis_type x_or_y );
   
void gfm_set_axis_crosses( enum axis_type x_or_y, GLfloat posit );

void gfm_set_initial_clip( GLfloat xlo, GLfloat xhi, GLfloat ylo, GLfloat yhi,
                           GLfloat zlo, GLfloat zhi );

void
gfm_set_padding( GLfloat xlo, GLfloat xhi, GLfloat ylo, GLfloat yhi,
                 GLfloat zlo, GLfloat zhi );

/* {{{ predefined color scheme */
/**
 * Function that defines a predefined color scheme.
 *
 * @param CS
 * @param NumC
 * @param Menu
 */
GLOBAL void gfmColorSchemePredefiningFunc(ColorScheme3f *CS[MAX_CS+1],
                                          int *NumC, int Menu);

/* }}} */

/* }}} */

/** @} */

#endif /* GLUT_FOR_MCMC_HFLAG */

/* {{{ Log */
/*
 * $Log: gldisp.h,v $
 * Revision 1.4  2005/08/22 23:56:39  myrna
 * do_auto_sweep() is now included in lm_auto.c;
 * prnt_locnam() and fmt_locnamlin() are included in Stuff/
 *
 * Revision 1.3  2005/08/14 00:19:36  thompson
 * Adding copyright statement at file head.
 *
 * Revision 1.2  2004/11/18 19:16:38  myrna
 * errsprogs.h:  Message added for program not compiled with display libraries
 * globs.h:      Current display iteration added
 * gldisp.h:     Include statement is corrected
 *
 * Revision 1.1  2004/11/16 00:30:09  myrna
 * Header files are added for the GLDisp/ library functions which provide
 * graphics rendering using OpenGL and freeGLUT.
 *
 * Revision 1.1.2.1  2003/06/16 18:55:31  nali
 * Added GFMCMC directory
 *
 * Revision 1.5  2002/09/10 00:26:32  nali
 * Changed function parameter names X and Y to xcoord and ycoord to
 * avoid name clash.
 *
 * Revision 1.4  2002/09/09 23:23:33  nali
 *
 * Fixed variable names N and Y which conflict with macros defined in
 * globs.h.
 *
 * Revision 1.3  2002/09/05 17:49:57  nali
 * doxyenizing ...
 *
 * Revision 1.2  2002/08/29 21:48:34  nali
 * doxyenizing...
 *
 */
/* vim: sw=3
 */
/* }}} */
#endif /* GLDISP_H */
