/******************************************************************************/
/*                                                                            */
/* The simped_fgl subprogram simulates patterns of identity by descent in a   */
/* given pedigree.                                                            */
/*                                                                            */
/******************************************************************************/

#define PNAM "ibd_create"

#include "parsie.h"
#include "rans.h"
#include "quantset.h"

#include "ibd_create.h"
#include "simped_fgl.h"

int simped_fgl (int argc, char **argv)
{
   FILE *oScorF = NULL;
   FILE *oSeedF;

   Gamete curr_gamete;                  /* Local copy of the current gamete.  */
   Breaklist *newbreaks;                /* New set of FGL change points.      */

   Person *indivs;                      /* Vector of pointers to all the      */
   /*                                      individuals read in from the named */
   /*                                      pedigree file.                     */

   Person mother;                       /* Current mother structure.          */
   Person father;                       /* Current father structure.          */

   long long chrlength;                 /* Chromosome length in micro         */
   /*                                      centimorgans.                      */

   int   cur_fgl;                       /* Current founder genome label.      */
   int   gam_ndx;                       /* Current gamete's subscript.        */
   int   ped_ndx;                       /* Pedigree and nghd[] subscript.     */
   int   gam_wch;                       /* 0 -> maternal, 1 -> paternal.      */

   /* ................................  SET-UP   ............................ */

   PCleanUp = clean_create;
   atexit (clean_morg);

   setupBasic ("PR", argc, argv, PNAM);

   /* Open the output scores file to use as simped_fgl's results file.        */

   if (oScorF == NULL)
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);

   /* The length of the chromosome has been stored in ChromLen as a double    */
   /* precision real number.  This needs to be converted into micro           */
   /* centimorgans, which is a long long integer.                             */

   chrlength = cent2micro (ChromLen);

   /* Allocate storage for a local vector in which we will store everything   */
   /* relating to the pedigree that this subprogram will need.                */

   indivs = (Person *) malloc (PedSize * sizeof (Person));
   alloc_chk (WHERE, indivs);

   /* Go through the nghd array to get information for each individual, then  */
   /* perform the simulation of IBD on the pedigree.                          */

   cur_fgl = 0;
   gam_ndx = 0;

   for (ped_ndx = 0; ped_ndx < PedSize; ++ped_ndx)
   {
      /* Determine whether or not this individual is a founder.               */

      if (isNullNode (nghd[ped_ndx].ma) && isNullNode (nghd[ped_ndx].pa))
      {
         /* If this individual is a founder:                                  */

         indivs[ped_ndx].maternal =
            (Chromseglist *) malloc (sizeof (Chromseglist));
         alloc_chk (WHERE, indivs[ped_ndx].maternal);
         indivs[ped_ndx].maternal->current =
            (Chromseg *) malloc (sizeof (Chromseg));
         alloc_chk (WHERE, indivs[ped_ndx].maternal->current);

         indivs[ped_ndx].maternal->current->fgl = ++cur_fgl;
         indivs[ped_ndx].maternal->current->endpos = chrlength;
         indivs[ped_ndx].maternal->next = NULL;

         indivs[ped_ndx].paternal =
            (Chromseglist *) malloc (sizeof (Chromseglist));
         alloc_chk (WHERE, indivs[ped_ndx].paternal);
         indivs[ped_ndx].paternal->current =
            (Chromseg *) malloc (sizeof (Chromseg));
         alloc_chk (WHERE, indivs[ped_ndx].paternal->current);

         indivs[ped_ndx].paternal->current->fgl = ++cur_fgl;
         indivs[ped_ndx].paternal->current->endpos = chrlength;
         indivs[ped_ndx].paternal->next = NULL;
      }
      else
      {
         /* If this individual is not a founder:                              */

         mother = indivs[nghd[ped_ndx].ma->index];
         father = indivs[nghd[ped_ndx].pa->index];

         newbreaks = makebreaks (chrlength, 1.0);
         indivs[ped_ndx].maternal = recombine (mother, newbreaks);
         free_breaks (newbreaks);

         newbreaks = makebreaks (chrlength, 1.0);
         indivs[ped_ndx].paternal = recombine (father, newbreaks);
         free_breaks (newbreaks);
      }

      if (DebugLevel >= 3)
      {
         /* For debugging purposes, print the contents of this individual's   */
         /* maternal and paternal gametes' FGL segment lists.                 */

         curr_gamete.name = nghd[ped_ndx].self;

         for (gam_wch = 0; gam_wch < 2; ++gam_wch)
         {
            curr_gamete.which = gam_wch;
            curr_gamete.index = gam_ndx++;
            if (gam_wch == 0)
               curr_gamete.seglist = indivs[ped_ndx].maternal;
            else
               curr_gamete.seglist = indivs[ped_ndx].paternal;

            prnt_gam_segs (curr_gamete);
         }
      }

      /* Print the resulting simulated IBD for the current individual.  Note  */
      /* that generation is set to 0 to match the simped_fgl output format.   */

      indivs[ped_ndx].sex = nghd[ped_ndx].sex % 2;

      for (gam_wch = 0; gam_wch < 2; ++gam_wch)
      {
         fprintf (oScorF, "%s 0 %s %s %d %d", nghd[ped_ndx].self,
                  nghd[ped_ndx].ma->self, nghd[ped_ndx].pa->self,
                  indivs[ped_ndx].sex, gam_wch);
         if (gam_wch == 0)
            printbreaks (oScorF, indivs[ped_ndx].maternal);
         else
            printbreaks (oScorF, indivs[ped_ndx].paternal);
      }
   }

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   if (oScorF)
      fclose (oScorF);

   /* We are done, so free up the storage that has been allocated by the      */
   /* simped_fgl subroutine.                                                  */

   free_generation (indivs, PedSize);

   return 0;
}
