/******************************************************************************/
/*                                                                            */
/* Program to do a basic analysis of IBD inference with no LD.                */
/*                                                                            */
/* It also includes options to simulate the ibd status of chromosomes.        */
/* We will make the chromosomes 2 Morgans, so most will have an IBD segment.  */
/*                                                                            */
/******************************************************************************/

#define DEFINE_DATA
#define PNAM "ibd_haplo"

#include "parsie.h"
#include "rans.h"
#include "markers.h"
#include "lsample.h"

#include "ibd_haplo_directory.h"
#include "ibd_create.h"
#include "ibd_haplo.h"

#include "compute_ibd_probs.c"
#include "hap_fwd_bk_compute.c"
#include "haplo_process.c"
#include "transitions.c"
#include "orbits.c"
#include "haplo_debug.c"

int main (int argc, char **argv)
{
   /* When the data are partially phased, we can use Morgan's extra input     */
   /* file for reading in the partial phasing data.  This will be the only    */
   /* situation where the iXtra FILE variable will be needed.                 */

   FILE *iXtra;
   FILE *oScorF;

   FILTAB_PTR oXtra_pft;
   FILE *oXtra = NULL;

   FILE *oSeedF;

   int   i1;                            /* Subscript of the first individual  */
   /*                                      in a pair of individuals.          */

   int   i2;                            /* Subscript of the second individual */
   /*                                      in a pair of individuals.          */

   int   ix;                            /* Subscript for current individual.  */
   int   m1;                            /* Subscript to current marker.       */

/******************************************************************************/
/*                                                                            */
/* The executable portion of the main program starts here.                    */
/*                                                                            */
/******************************************************************************/

   /* ................................  SET-UP   ............................ */

   PCleanUp = clean_ibd;
   atexit (clean_morg);

   setupBasic ("AR", argc, argv, PNAM);

   /* Only call set_ProbIx if not running an all pairs case.                  */

   if (DoingAllPairs)
   {
      /* Call alloc_for_pairs to allocate storage for all pairs case.         */
      alloc_for_pairs ();

      /* There are two variables that need to be initialized for use by the   */
      /* set_for_pair subroutine.  Initialize these variables here.           */

      indiv_names = ObsNam;
      num_names = NObs;
   }
   else
   {
      /* Call set_ProbIx to save the needed indices in ProbIx.                */
      set_ProbIx ();
   }

   if (PhaseofData == PARTIALPHASE)
   {
      /* Open the extra input file to read in the partial phasing data.       */

      if (!(iXtra = open_fil (XTRA)))
         give_up (ERR_qit);

      phas_nam = alloc_char2d (NObs, MAXNAMLEN + 1);
      phasdatindiv = alloc_int2d (NObs, NMarks);

      /* Read in one line of phasing data for each individual's pair of       */
      /* haplotypes.                                                          */

      for (ix = 0; ix < NObs; ++ix)
      {
         fscanf (iXtra, "%s %*s", phas_nam[ix]);
         if (strncmp (ObsNam[ix], phas_nam[ix], MAXNAMLEN) != 0)
         {
            fprintf (stderr, " A name, \"%s\", read from the", phas_nam[ix]);
            fprintf (stderr, " phasing data file does not match the");
            fprintf (stderr, " corresponding name, \"%s\", from", ObsNam[ix]);
            fprintf (stderr, " the marker data file  (W)\n\n");
         }

         for (m1 = 0; m1 < NMarks; ++m1)
            fscanf (iXtra, "%d", &(phasdatindiv[ix][m1]));
      }

      /* Close the extra input file since we are done reading in the partial  */
      /* phasing data.                                                        */

      fclose (iXtra);
   }

   /* Copy the allele frequencies from the global variable, Qq, to the local  */
   /* variable, allefreq[].                                                   */

   allefreq = alloc_dbl1d (NMarks);
   for (m1 = 0; m1 < NMarks; ++m1)
      allefreq[m1] = Qq[m1 + 1][1];

   /* Compute the positions of all the markers involved in the analysis.      */

   snppos = alloc_dbl1d (NMarks);
   snppos[0] = FirstMarkPos;
   for (m1 = 1; m1 < NMarks; ++m1)
      snppos[m1] = snppos[m1 - 1] + Mdist[m1 - 1];

   /* Open the output scores file.                                            */

   if (!(oScorF = open_fil (OSCORE)))
      give_up (ERR_qit);

   /* If the user has named an output extra file, open it now.                */

   oXtra_pft = get_filent (OXTR);
   if (oXtra_pft != NULL && oXtra_pft->fname)
      if (!(oXtra = open_fil (OXTR)))
         give_up (ERR_qit);

   fprintf (stdout, "\n");

   /* Call init_ibd_haplo to initialize those constructs specific to each     */
   /* score set size.  All of the pointers to these contructs are saved in    */
   /* score_set_size_pointers.  There are also a couple of one time only      */
   /* subroutines called at the end of init_ibd_haplo.                        */

   init_ibd_haplo ();

   /* The compute_ibd_probs subroutine is called to execute the main loop in  */
   /* ibd_haplo.  In this subroutine the IBD probabilities for each score set */
   /* indicated by the ibd_haplo user are computed.                           */

   if (DoingAllPairs)
   {
      for (i1 = 0; i1 < NObs; ++i1)
      {
         for (i2 = (i1 + 1); i2 < NObs; ++i2)
         {
            /* Call the set_for_pair subroutine to set up the global proband  */
            /* gamete variables for the given pair of individuals.            */

            set_for_pair (i1, i2);

            compute_ibd_probs (i1, i2, oScorF, oXtra);
            if (i1 < NObs - 2)
               fprintf (stdout, "\n");

            ++ScSetNo[0][0];
         }
      }
   }
   else
      compute_ibd_probs (UNDEF, UNDEF, oScorF, oXtra);

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   /* Close the scores file and output extra file.                            */

   fflush (oScorF);
   fclose (oScorF);

   if (oXtra)
   {
      fflush (oXtra);
      fclose (oXtra);
   }

   exit (EXIT_SUCCESS);
}

/******************************************************************************/
/*                                                                            */
/* This subroutine sets the values in the ProbIx array.  It stores the index  */
/* for each proband gamete to the named individual's marker data in ObsMark.  */
/* If a proband gamete name cannot be found in ObsNam, this subroutine prints */
/* a fatal error message before terminating the program.                      */
/*                                                                            */
/******************************************************************************/

void set_ProbIx (void)
{
   int   terminate;                     /* 0 -> All proband gamete names were */
   /*                                           found, so continue program.   */
   /*                                      1 -> One or more names not found,  */
   /*                                           so terminate the program.     */

   int   num_found;                     /* Number of named proband gametes    */
   /*                                      found in the marker data file.     */

   int   ind_ndx;                       /* Subscript for proband gamete.      */
   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to current score set.    */
   int   g1;                            /* Subscript to this proband gamete.  */
   int   ga;                            /* Subscript to alternate gamete.     */
   int   ix;                            /* Subscript to current individual.   */

   /* Start with the assumption that there won't be any names not found.      */
   terminate = FALSE;

   /* Initialize the number of named proband gametes found in the marker data */
   /* file to 0.                                                              */

   num_found = 0;

   /* We cannot continue executing code in the set_ProbIx subroutine unless   */
   /* there is at least one "set proband gametes ..." parameter statement.    */

   if (NScSet == NULL)
      give_up (ERA_spg);

   for (s1 = 0; s1 < NScSet[c1]; ++s1)
   {
      for (g1 = 0; g1 < NPrGam[c1][s1]; ++g1)
      {
         /* Initialize ind_ndx to INVALID_INDEX so that we can determine if   */
         /* the gamete's name was found among the names read in from the      */
         /* marker data file.                                                 */

         ind_ndx = INVALID_INDEX;

         for (ix = 0; ix < NObs; ++ix)
         {
            if (strncmp (ProbNam[c1][s1][g1], ObsNam[ix], MAXNAMLEN) == 0)
            {
               ProbIx[c1][s1][g1] = ind_ndx = ix;
               ++num_found;
               break;
            }
         }

         if (ind_ndx == INVALID_INDEX)
         {
            give_err (ERR_mpg, ProbNam[c1][s1][g1]);
            terminate = TRUE;
         }
      }

      /* If the data are not being analyzed as phased data we need to check   */
      /* that the proband gametes are only for sets of whole individuals.     */

      if (PhaseofData != PHASED)
      {
         for (g1 = 0; g1 < NPrGam[c1][s1]; ++g1)
         {
            /* The following statement computes the subscript to the proband  */
            /* gamete which should be paired with the current proband gamete. */
            /* In other words, if g1 is even ga = g1 + 1, and if g1 is odd    */
            /* ga = g1 - 1.                                                   */

            ga = ((g1 / 2) * 2 == g1) ? g1 + 1 : g1 - 1;

            if (ga == NPrGam[c1][s1]
                || strcmp (ProbNam[c1][s1][g1], ProbNam[c1][s1][ga]))
            {
               give_err (ERR_nwi, ScSetNo[c1][s1], ProbNam[c1][s1][g1]);
               terminate = TRUE;
            }
         }
      }
   }

   if (num_found == 0)                  /* No named proband gametes found?    */
      give_up (ERA_npg);

   if (terminate)
      give_up (ERR_qit);

   return;
}

/******************************************************************************/
/*                                                                            */
/* This subroutine initializes those constructs which differ depending on the */
/* score set sizes needed to compute the IBD probabilities for the score sets */
/* the user has indicated in their Morgan parameter statements.               */
/*                                                                            */
/******************************************************************************/

void init_ibd_haplo (void)
{
   double tp;                           /* Used to normalize probabilities.   */

   int  *printed;                       /* Vector used to remember whether or */
   /*                                      not output has already been done   */
   /*                                      for each score set.                */

   int   ss_size;                       /* Size of the current score set.     */
   int   ss_size_ndx;                   /* Current score set size subscript.  */

   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to score set.            */

   int   i1, i2;                        /* Subscripts to use in for loops.    */

/******************************************************************************/
/*                                                                            */
/* The executable portion of this subroutine starts here.                     */
/*                                                                            */
/******************************************************************************/

   /* Construct those marginal IBD probabilities vectors and transition       */
   /* matrices which we will need for the proband gamete score sets the       */
   /* ibd_haplo user has indicated.                                           */

   score_set_size_pointers = malloc (MxGam * sizeof (SS_SIZE_STRUCTURE));
   alloc_chk (WHERE, score_set_size_pointers);
   memset (score_set_size_pointers, 0, MxGam * sizeof (SS_SIZE_STRUCTURE));

   for (s1 = 0; s1 < NScSet[c1]; ++s1)
   {
      ss_size = NPrGam[c1][s1];         /* Get size of the current score set. */
      ss_size_ndx = ss_size - 1;        /* Set the score set size subscript.  */

      if (score_set_size_pointers[ss_size_ndx].state_matrix == NULL)
      {
         /* So long as it has not previously been done, call the              */
         /* get_ibd_state_mat subroutine to obtain a matrix of IBD states for */
         /* the current score set size.  This subroutine also has to compute  */
         /* a vector of 0's and 1's reflecting where there are duplicate      */
         /* rows.  Since the set_tmat_2011 subroutine will also need this     */
         /* vector later, it will be retained as the duplicated_rows vector.  */

         get_ibd_state_mat (ss_size, &num_states, &duplicated_rows,
                            &state_matrix);
         score_set_size_pointers[ss_size_ndx].state_matrix = state_matrix;

         /* Construct the genotypic averaging orbits data matrix unless the   */
         /* data are phased.                                                  */

         if (PhaseofData != PHASED)
            makeOrbits (ss_size, num_states, state_matrix);
         else
         {
            /* Just make sure that the orbits matrix and the orbitSize vector */
            /* variables are initialized to something valid such as NULL.     */

            orbits = NULL;
            orbitSizes = NULL;
         }

         score_set_size_pointers[ss_size_ndx].num_orbits = num_orbits;
         score_set_size_pointers[ss_size_ndx].orbitSizes = orbitSizes;
         score_set_size_pointers[ss_size_ndx].orbits = orbits;

         /* Compute the vector of marginal ibd probabilities for the current  */
         /* score set size.                                                   */

         marg_probs = alloc_dbl1d (num_states);
         set_marg_ibd (ss_size, num_states, marg_probs, state_matrix);
         score_set_size_pointers[ss_size_ndx].marg_probs = marg_probs;

         if (num_states > 2)
         {
            /* Compute tmat for this score set's number of haplotypes.  The   */
            /* memory for tmat is allocated in the set_tmat_20xx subroutines. */

            if (StateTransMat == STM_2011)
               set_tmat_2011 (ss_size, PopKinship, &tmat);
            else
               set_tmat_2009 ();

            /* Convert the continous process rate matrix into the jump chain  */
            /* matrix.                                                        */

            tmat_rate = alloc_dbl1d (num_states);

            for (i1 = 0; i1 < num_states; ++i1)
               tmat_rate[i1] = -tmat[i1][i1];

            tp = 0.0;
            for (i1 = 0; i1 < num_states; ++i1)
               tp += tmat_rate[i1] * marg_probs[i1];

            for (i1 = 0; i1 < num_states; ++i1)
            {
               for (i2 = 0; i2 < num_states; ++i2)
                  tmat[i1][i2] = tmat[i1][i2] / tmat_rate[i1];
               tmat[i1][i1] = 0.0;
            }

            for (i1 = 0; i1 < num_states; ++i1)
            {
               for (i2 = 0; i2 < num_states; ++i2)
                  tmat[i1][i2] = ((1.0 - TransMatNullFr) * tmat[i1][i2]) +
                     (TransMatNullFr * tmat_rate[i2] * marg_probs[i2] / tp);
            }

            score_set_size_pointers[ss_size_ndx].tmat_rate = tmat_rate;
            score_set_size_pointers[ss_size_ndx].tmat = tmat;
         }

         free_1d ((void *) &duplicated_rows);

         /* Compute 1 cM transmat for this score set's number of haplotypes.  */

         transmat = alloc_dbl2d (num_states, num_states);
         set_transmat (ss_size, num_states, 1.0, marg_probs, transmat);
         score_set_size_pointers[ss_size_ndx].transmat = transmat;
      }
   }

   if (PrintLevel >= 3)
   {
      /* Print the equilibrium state probabilities and the state transition   */
      /* matrix once for each score set size indicated by the "set proband    */
      /* gametes" parameter statements.                                       */

      /* When lexicographical output order is not wanted and if the number of */
      /* haplotypes in the current score set is 4, we still want to reorder   */
      /* the lexicographically ordered vector entries into the "traditional"  */
      /* (ie, Jacquard) output order previously used.                         */

      printed = alloc_int1d (MxGam);    /* Initialized to all 0's.            */

      for (ss_size = 2; ss_size <= MxGam; ++ss_size)
      {
         ss_size_ndx = ss_size - 1;     /* Set the score set size subscript.  */
         num_states = PTgg[ss_size];    /* Number of states for score set.    */

         if (printed[ss_size_ndx] == FALSE)
         {
            marg_probs = score_set_size_pointers[ss_size_ndx].marg_probs;
            if (marg_probs != NULL)
            {
               if (Lexico == N && ss_size == 4 && DebugLevel != 3)
               {
                  printf (" The traditionally ordered %d", ss_size);
                  printf ("-gamete equilibrium state probabilities are:\n");
                  for (i1 = 0; i1 < num_states; ++i1)
                     printf (" %7.5f", marg_probs[lex2gen_ph[i1]]);
               }
               else
               {
                  printf (" The lexicographically ordered %d", ss_size);
                  printf ("-gamete equilibrium state probabilities are:\n");
                  for (i1 = 0; i1 < num_states; ++i1)
                     printf (" %7.5f", marg_probs[i1]);
               }

               printf ("\n\n");
               fflush (stdout);
            }

            tmat_rate = score_set_size_pointers[ss_size_ndx].tmat_rate;
            if (tmat_rate != NULL)
            {
               if (Lexico == N && ss_size == 4 && DebugLevel != 3)
               {
                  printf (" The traditionally ordered %d", ss_size);
                  printf ("-gamete state departure rates are:\n");
                  for (i1 = 0; i1 < num_states; ++i1)
                     printf (" %7.5f", tmat_rate[lex2gen_ph[i1]]);
               }
               else
               {
                  printf (" The lexicographically ordered %d", ss_size);
                  printf ("-gamete state departure rates are:\n");
                  for (i1 = 0; i1 < num_states; ++i1)
                     printf (" %7.5f", tmat_rate[i1]);
               }

               printf ("\n\n");
               fflush (stdout);
            }

            transmat = score_set_size_pointers[ss_size_ndx].transmat;
            if (transmat != NULL)
            {
               if (Lexico == N && ss_size == 4 && DebugLevel != 3)
               {
                  printf (" At 1 cM, the traditionally ordered");
                  printf (" %d-gamete state transition matrix is:\n", ss_size);
                  for (i1 = 0; i1 < num_states; ++i1)
                  {
                     for (i2 = 0; i2 < num_states; ++i2)
                     {
                        printf (" %7.5f",
                                transmat[lex2gen_ph[i1]][lex2gen_ph[i2]]);
                     }
                     printf ("\n");
                  }
               }
               else
               {
                  printf (" At 1 cM, the lexicographically ordered");
                  printf (" %d-gamete state transition matrix is:\n", ss_size);
                  for (i1 = 0; i1 < num_states; ++i1)
                  {
                     for (i2 = 0; i2 < num_states; ++i2)
                     {
                        printf (" %7.5f", transmat[i1][i2]);
                     }
                     printf ("\n");
                  }
               }

               printf ("\n\n");
               fflush (stdout);
            }

            printed[ss_size_ndx] = TRUE;
         }
      }

      free_1d ((void *) &printed);
   }

   /* The init_ibd_lengths subroutine initializes the four static integer     */
   /* variables where selected output field widths used by compute_ibd_probs  */
   /* are saved.                                                              */

   init_ibd_lengths ();

   /* The setup_out2sel_ndx subroutine initializes the out2sel_ndx vector of  */
   /* subscripts used to map the marker numbers in MarOut to where they are   */
   /* located in MarSel.                                                      */

   setup_out2sel_ndx ();

   return;
}

/******************************************************************************/
/*                                                                            */
/* The setup_out2sel_ndx subroutine should only be called a single time.  It  */
/* is called to allocate and setup the out2sel_ndx vector.  Each integer in   */
/* the out2sel_ndx will be the subscript into MarSel where the corresponding  */
/* marker number in MarOut is also located (in MarSel).                       */
/*                                                                            */
/******************************************************************************/

void setup_out2sel_ndx (void)
{
   int   marsel_ndx;                    /* Current subscript into MarSel.     */
   int   marout_ndx;                    /* Current subscript into MarOut.     */

   int   cur_mark_num;                  /* Current marker number.             */

   if (out2sel_ndx == NULL)
   {
      out2sel_ndx = alloc_int1d (NMarOut);
      marsel_ndx = 0;

      for (marout_ndx = 0; marout_ndx < NMarOut; ++marout_ndx)
      {
         cur_mark_num = MarOut[marout_ndx];

         while (MarSel[marsel_ndx] < cur_mark_num)
            ++marsel_ndx;

         if (MarSel[marsel_ndx] == cur_mark_num)
            out2sel_ndx[marout_ndx] = marsel_ndx;

         else if (MarSel[marsel_ndx] > cur_mark_num)
         {
            /* As long as both the MarSel and MarOut integer vectors are in   */
            /* ascending numerical order and every marker number present in   */
            /* MarOut is also in MarSel, then MarSel[marsel_ndx] should never */
            /* be greater than the current marker number found in MarOut.  So */
            /* we need a terminal error message if this should ever occur.    */

            fprintf (stderr, "\n");
            fprintf (stderr, " Terminal error: marker number %d", cur_mark_num);
            fprintf (stderr, " is in MarOut but not in MarSel.\n");
            exit (EXIT_FAILURE);
         }
      }
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_ibd_haplo subroutine should be called while exiting from the      */
/* ibd_haplo main program.  It frees up all the storage which was allocated   */
/* (but not freed) during the execution of the ibd_haplo main program.        */
/*                                                                            */
/******************************************************************************/

void free_ibd_haplo (void)
{
   int   g1;                            /* Subscript to current gamete.       */

   free_1d ((void *) &allefreq);
   free_1d ((void *) &snppos);
   free_1d ((void *) &out2sel_ndx);

   for (g1 = 0; g1 < MxGam; ++g1)
   {
      if (score_set_size_pointers[g1].state_matrix != NULL)
         free_2d ((void *) &score_set_size_pointers[g1].state_matrix);
      if (score_set_size_pointers[g1].orbitSizes != NULL)
         free_1d ((void *) &score_set_size_pointers[g1].orbitSizes);
      if (score_set_size_pointers[g1].orbits != NULL)
         free_2d ((void *) &score_set_size_pointers[g1].orbits);
      if (score_set_size_pointers[g1].marg_probs != NULL)
         free_1d ((void *) &score_set_size_pointers[g1].marg_probs);
      if (score_set_size_pointers[g1].tmat_rate != NULL)
         free_1d ((void *) &score_set_size_pointers[g1].tmat_rate);
      if (score_set_size_pointers[g1].tmat != NULL)
         free_2d ((void *) &score_set_size_pointers[g1].tmat);
      if (score_set_size_pointers[g1].transmat != NULL)
         free_2d ((void *) &score_set_size_pointers[g1].transmat);
   }
   free_1d ((void *) &score_set_size_pointers);

   if (PhaseofData == PARTIALPHASE)
   {
      free_2d ((void *) &phas_nam);
      free_2d ((void *) &phasdatindiv);
   }

   return;
}
