/* Program to do a basic analysis of IBD inference with no LD.                */
/* Up to 4 genomes, or genotypic pairs.                                       */

/* Also includes options to simulate the ibd status of chromosomes.           */
/* Initially these will just be segregants from first cousins.                */
/* We will make the chromosomes 2 Morgans, so most have an IBD segment.       */

#define DEFINE_DATA
#define PNAM "ibd_haplo"

#include "parsie.h"
#include "rans.h"
#include "markers.h"
#include "ibd_haplo.h"

#include "haplo_process.c"
#include "hap_fwd_bk_compute.c"
#include "transitions.c"
#include "joint.c"
#define JOINT 1
#define DEBUG 1

int main (int argc, char *argv[])
{
   /* When the data are partially phased, we can use Morgan's extra input     */
   /* file for reading in the partial phasing data.  This will be the only    */
   /* situation where the iXtra FILE variable will be needed.                 */

   FILE *oXtra, *iXtra;

   double *snppos;                      /* Selected marker positions.         */
   double *allefreq;                    /* Allele frequencies at each marker. */

   double **qq_ibd_final;
   double **qq_ibd;
   double **pprob;
   int  *simStates;

   double *ffibd;                       /* Overall inbr,  Marginal ibd prob.  */

   char  temp_buff[8];                  /* A temporary buffer to format the   */
   /*                                      highest selected marker number to. */

   /* Arrays of phasing data.                                                 */
   int **phasdatindiv;                  /* First is for individuals.          */
   int **phasdat;                       /* Second is for pairs of individuals. */

   int  *phasptr = NULL;

   int   ns, nph, nstmp;                /* Number of states, and phenotypes.  */
   int   nhap;                          /* Number of haplos, and start and    */
   /*                                      end of SNP block.                  */

   int **haplodat;                      /* Haplotypic data.                   */
   int  *hapdat;

   int   indiv;                         /* Subscript for current individual.  */

   int   tmp1, tmp2;                    /* The two tmp variables are holders  */
   /*                                      for swapping alleles.              */

   int   nc;                            /* Character length of longest name.  */
   int   nm;                            /* Character length of highest        */
   /*                                      numbered selected marker.          */

   int   half_hss;                      /* HapSetSize / 2.                    */
   int   hs_ndx;                        /* Current offset into ObsNam for the */
   /*                                      current haplotype set.             */

   int   i, j, k;

   /* things added for joint sampling */
   unsigned int *allowedstates;         /* array of state bitmasks */
   int   NIndivs;
   int  *mainOrd, *inverseMainOrd, *subOrd;
   /* Used for random reordering */
   double *sortRands;                   /* in joint sampling */
   int ***ibdStates;

   int ***genomeLabels;
   int   counter, NPairs, pairsCounter, iter, NIter;
   /***************/
/******************************************************************************/
/*                                                                            */
/* The executable portion of the main program starts here.                    */
/*                                                                            */
/******************************************************************************/

   initMorgProg (argv);

   /* Parse input parameter statements.                                       */
   if (!proc_pars (argc, argv, PNAM))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* Check statements for completeness and consistency.                      */
   /* Set up global parameters and data structures.                           */
   if (!proc_auto ())
      exit (EXIT_FAILURE);

   prntStatusCP ();

   /* Call the get_ibd_state_mat subroutine to obtain the matrix of IBD       */
   /* states.  This subroutine also has to compute a vector of 0's and 1's    */
   /* reflecting where there are duplicate rows.  Since the createmat         */
   /* subroutine will also need this vector later, it will be retained as     */
   /* the duplicated_rows vector.                                             */

   get_ibd_state_mat (HapSetSize, &num_states, &duplicated_rows, &state_matrix);

   ns = min (num_states, 203);          /* Limit ns to not greater than 203.   */
   /* this is # of states for 6-haplotypes analysis */

   nph = 1 << HapSetSize;               /* # of possible phens is 2^HapSetSize */

   skipmark = 1;                        /* what is this for? do we still use it? */

   nhap = NumberHapSets * HapSetSize;
   NIndivs = nhap / 2;

   /* Allocate storage for the numerous arrays we will need.                  */

   snppos = malloc (NMarks * sizeof (double));
   haplodat = alloc_int2d (nhap, NMarks);

   hapdat = malloc (NMarks * sizeof (int)); /* All cases: analysis data.      */
   allefreq = malloc (NMarks * sizeof (double));

   transmat = alloc_dbl2d (ns, ns);
   qq_ibd = alloc_dbl2d (NMarks, ns);
   pprob = alloc_dbl2d (NMarks, ns);

   /* Copy the genotypic data from the ObsMark[][][] array.                   */

   for (i = 0; i < nhap; i = i + 2)
   {
      indiv = i / 2;
      for (j = 0; j < NMarks; ++j)
      {
         haplodat[i][j] = ObsMark[indiv][j][0];
         haplodat[i + 1][j] = ObsMark[indiv][j][1];
      }
   }

   /* Initialize the integer variable, nc, to the number of characters in the */
   /* longest name of an observed individual in the dataset.                  */

   nc = 0;
   if (NObs > 0)
   {
      for (i = 0; i < NObs; ++i)
      {
         j = strlen (ObsNam[i]);
         nc = j > nc ? j : nc;
      }
   }

   /* To make the data written to the extra output file line up nicely, find  */
   /* out how many characters are needed to write the marker number of the    */
   /* highest numbered selected marker.                                       */

   sprintf (temp_buff, "%d", MarSel[NMarks - 1]);
   nm = strlen (temp_buff);

   if (PhaseofData == PARTIALPHASE)
   {
      /* One line of phasing data per haplotypic pair.                        */
      phasdatindiv = alloc_int2d (nhap / 2, NMarks);

      /* Could ditch this matrix and just compute the vector for each compute run. */
      phasdat = alloc_int2d (NumberHapSets, NMarks);

      /* Open the extra input file to read in the partial phasing data.       */

      if (!(iXtra = open_fil (XTRA)))
         give_up (ERR_qit);

      printf (" Reading phasing status data:");

      for (i = 0; i < nhap / 2; ++i)
      {
         fscanf (iXtra, "%*d %*d");
         for (j = 0; j < NMarks; ++j)
         {
            fscanf (iXtra, "%d", &(phasdatindiv[i][j]));
            if (phasdatindiv[i][j] == DISCORDANT)
            {
               tmp1 = haplodat[2 * i][j];
               tmp2 = haplodat[2 * i + 1][j];
               haplodat[2 * i][j] = tmp2;
               haplodat[2 * i + 1][j] = tmp1;
               phasdatindiv[i][j] = 1;
            }
         }
      }

      /* Close the extra input file since we are done reading in the partial  */
      /* phasing data.                                                        */

      fclose (iXtra);

      for (i = 0; i < NumberHapSets; ++i)
      {
         for (j = 0; j < NMarks; ++j)
         {
            /* It's only phased for the HMM model if BOTH are phased.         */
            phasdat[i][j] = phasdatindiv[2 * i][j] * phasdatindiv[2 * i + 1][j];
         }
      }

      for (i = 0; i < nhap / 2; ++i)
      {
         printf ("\n Individual name %*s:", nc, ObsNam[i]);
         for (j = 0; j < 10; ++j)
            printf (" %d", phasdatindiv[i][j]);
      }

      for (i = 0; i < NumberHapSets; ++i)
      {
         printf ("\n Pair (4-haplotypes) %*s %*s:", nc, ObsNam[2 * i], nc,
                 ObsNam[2 * i + 1]);
         for (j = 0; j < 10; ++j)
            printf (" %d", phasdat[i][j]);
      }

      printf ("\n\n\n");
      fflush (stdout);
   }

   /* Compute the positions of all the markers involved in the analysis.      */

   snppos[0] = FirstMarkPos;
   for (j = 1; j < NMarks; ++j)
      snppos[j] = snppos[j - 1] + Mdist[j - 1];

   /* Copy the allele frequencies from the global variable, Qq, to the local  */
   /* variable, allefreq[].                                                   */

   for (j = 0; j < NMarks; ++j)
      allefreq[j] = Qq[j + 1][1];

   /* Open the extra output file.                                             */

   if (!(oXtra = open_fil (OXTR)))
      give_up (ERR_qit);

   /* Make the data vector of states from the haplotypic data.                */
   /* First, haplodat has switched from 0,1 to 1,2 (with 0 missing).          */
   /* Second, we can now do several groups of 4 haplotypes: for 1 set kk = 0. */

   hh = 1 - PopKinship;
   gg = PopKinship;

   ffibd = malloc (ns * sizeof (double));
   set_marg_ibd (HapSetSize, ns, ffibd, state_matrix);
   printf (" Equilib state probs at kinship %f\n", PopKinship);
   for (i = 0; i < ns; ++i)
      printf (" %5.3f", ffibd[i]);
   printf ("\n");
   fflush (stdout);

   if (JOINT)
   {
      mainOrd = malloc (NIndivs * sizeof (int));
      inverseMainOrd = malloc (NIndivs * sizeof (int));
      subOrd = malloc (NIndivs * sizeof (int));
      sortRands = malloc (NIndivs * sizeof (double));
      ibdStates = malloc (NIndivs * sizeof (int **));

      NPairs = NIndivs * (NIndivs - 1) / 2;
      NIter = 1;

      for (i = 0; i < NIndivs; ++i)
      {
         ibdStates[i] = malloc (NIndivs * sizeof (int *));
         for (j = 0; j < NIndivs; ++j)
         {
            if (j > i)
            {
               ibdStates[i][j] = malloc (NMarks * sizeof (int));
            }
            else
            {
               ibdStates[i][j] = NULL;
            }
         }
      }

      allowedstates = malloc (NMarks * sizeof (unsigned int));
      genomeLabels = malloc (NIndivs * sizeof (int **));
      for (i = 0; i < NIndivs; i++)
      {
         genomeLabels[i] = alloc_int2d (2, NMarks);
      }

      for (iter = 0; iter < NIter; iter++)
      {
         pairsCounter = 0;
         for (i = 0; i < NIndivs; ++i)
         {
            sortRands[i] = rans ();
         }
         tagsort (sortRands, mainOrd, NIndivs);
         for (i = 0; i < NIndivs; ++i)
         {
            inverseMainOrd[mainOrd[i]] = i;
         }

         setAllStates (allowedstates, NMarks);
         makeDataIndexed (hapdat, haplodat, mainOrd[0], mainOrd[1], NMarks);
         hap_fwd_compute_cond (hapdat, NMarks, nph, ns, transmat,
                               allefreq, ffibd, Mdist, pprob, PhaseofData,
                               phasptr, allowedstates);
         hap_sample_cond (NMarks, ns, ibdStates[0][1], transmat, ffibd,
                          Mdist, pprob, allowedstates);
         printf ("\n");
         progBar (iter + 1, NIter, (pairsCounter++) / (double) NPairs);
         //       printf ("\nsampling indivs %d and %d", inverseMainOrd[0],inverseMainOrd[1]);
         for (i = 2; i < NIndivs; i++)
         {
            for (j = 0; j < i; ++j)
            {
               sortRands[j] = rans ();
            }
            tagsort (sortRands, subOrd, i);

            /* first attach the first indiv (after reordering) 
             * to indiv i. A is subOrd[0], B is subOrd[1],
             * C is i */
            setAllStates (allowedstates, NMarks);
            arrayAB =
               ibdStates[MIN (subOrd[0], subOrd[1])][MAX
                                                     (subOrd[0], subOrd[1])];
            setAutozygStates (arrayAB, allowedstates, subOrd[0] > subOrd[1],
                              NMarks);
            makeDataIndexed (hapdat, haplodat, inverseMainOrd[subOrd[0]],
                             inverseMainOrd[i], NMarks);
            hap_fwd_compute_cond (hapdat, NMarks, nph, ns, transmat, allefreq,
                                  ffibd, Mdist, pprob, PhaseofData, phasptr,
                                  allowedstates);
            //           printf ("\nsampling indivs %d and %d", inverseMainOrd[subOrd[0]],inverseMainOrd[i]);
            progBar (iter + 1, NIter, (pairsCounter++) / (double) NPairs);
            hap_sample_cond (NMarks, ns, ibdStates[subOrd[0]][i], transmat,
                             ffibd, Mdist, pprob, allowedstates);

            /* now attach each remaining indiv j to indiv i,
             * conditioning on trio formed w/ each previous indiv. 
             * A is j, B is k, C is i */
            for (j = 1; j < i; ++j)
            {
               setAllStates (allowedstates, NMarks);
               for (k = 0; k < j; ++k)
               {
                  arrayAB =
                     ibdStates[MIN (subOrd[j], subOrd[k])][MAX
                                                           (subOrd[j],
                                                            subOrd[k])];
                  arrayBC = ibdStates[subOrd[k]][i];
                  setTrioStates (arrayAB, arrayBC, allowedstates,
                                 subOrd[j] > subOrd[k], 0, NMarks);
               }
               makeDataIndexed (hapdat, haplodat, inverseMainOrd[subOrd[j]],
                                inverseMainOrd[i], NMarks);
               hap_fwd_compute_cond (hapdat, NMarks, nph, ns, transmat,
                                     allefreq, ffibd, Mdist, pprob, PhaseofData,
                                     phasptr, allowedstates);
               //           printf ("\nsampling indivs %d and %d", inverseMainOrd[subOrd[j]],inverseMainOrd[i]);
               progBar (iter + 1, NIter, (pairsCounter++) / (double) NPairs);
               hap_sample_cond (NMarks, ns, ibdStates[subOrd[j]][i], transmat,
                                ffibd, Mdist, pprob, allowedstates);
            }
         }

         for (i = 0; i < NIndivs; i++)
         {
            for (j = 0; j < 2; j++)
            {
               for (k = 0; k < NMarks; k++)
               {
                  genomeLabels[i][j][k] = 2 * i + j + 1;
               }
            }
         }
         for (i = 0; i < NIndivs; i++)
         {
            if (i < (NIndivs - 1))
            {
               setAutozygosityLabels (ibdStates[i][i + 1],
                                      genomeLabels[inverseMainOrd[i]], 0,
                                      NMarks);
            }
            else
            {
               setAutozygosityLabels (ibdStates[i - 1][i],
                                      genomeLabels[inverseMainOrd[i]], 1,
                                      NMarks);
            }
            for (j = (i + 1); j < NIndivs; j++)
            {
               setIbdLabels (ibdStates[i][j], genomeLabels[inverseMainOrd[i]],
                             genomeLabels[inverseMainOrd[j]], NMarks);
            }
         }
         /* print out output */
         for (i = 0; i < NIndivs; i++)
         {
            if (DEBUG)
            {
               /* print out all pairwise IBD states */
               int   kk;
               for (j = i + 1; j < NIndivs; j++)
               {
                  for (k = j + 1; k < NIndivs; k++)
                  {
                     for (kk = 0; kk < NMarks; kk++)
                     {
                        if (!IS_SET
                            (trioMasks[ibdStates[i][j][kk]]
                             [ibdStates[j][k][kk]], ibdStates[i][k][kk]))
                           printf ("Aaaaagh %d %d %d %d\n", i, j, k, kk);
                     }
                  }
                  fprintf (oXtra, "%d %d ", i + 1, j + 1);
                  for (k = 0; k < NMarks; k++)
                  {
                     fprintf (oXtra, "%d ", ibdStates[i][j][k]);
                  }
                  fprintf (oXtra, "\n");
               }
            }
            else
            {
               /* print an IBD graph */
               counter = 0;
               for (j = 0; j < 2; j++)
               {
                  for (k = 1; k < NMarks; k++)
                  {
                     if (genomeLabels[i][j][k] != genomeLabels[i][j][k - 1])
                        counter++;
                  }
                  fprintf (oXtra, "%s %d %d %d", ObsNam[i], 0,
                           genomeLabels[i][j][0], counter);
                  for (k = 1; k < NMarks; k++)
                     if (genomeLabels[i][j][k] != genomeLabels[i][j][k - 1])
                        fprintf (oXtra, " %s %d ", ObsNam[2 * i + j],
                                 genomeLabels[i][j][k]);
                  fprintf (oXtra, "\n");
               }
            }
         }
      }
   }
   else
   {
      for (k = 0; k < NumberHapSets; ++k) /* k'th set of haplotypes.            */
      {
         printf ("\n Data vector for %d markers in set %d to be set", NMarks,
                 k + 1);
         fflush (stdout);
         make_data (hapdat, haplodat, HapSetSize, k, NMarks, 1, 0);

         printf ("\n Data vector for %d markers in set %d is set", NMarks,
                 k + 1);
         fflush (stdout);

         set_transmat (ns, transmat, ffibd, 1.0); /* Check transmat.             */
         for (i = 0; i < ns; ++i)
         {
            printf ("\n");
            for (j = 0; j < ns; ++j)
               printf (" %7.5f", transmat[i][j]);
         }

         printf ("\n Markers %d, phen %d , states %d", NMarks, nph, ns);

         /* If using the matrix of phased data, set the relevant pointer.        */
         if (PhaseofData == PARTIALPHASE)
            phasptr = phasdat[k];

         hap_fwd_compute (hapdat, NMarks, nph, ns, transmat, allefreq,
                          ffibd, Mdist, pprob, PhaseofData, phasptr,
                          HapSetSize);
         if (!DebugGLDisp)
         {
            hap_bck_compute (hapdat, NMarks, nph, ns, qq_ibd, transmat,
                             allefreq, ffibd, Mdist, pprob, PhaseofData,
                             phasptr, HapSetSize);
         }

         /* The output written to both the standard output file and qibd file    */
         /* are prefixed by printing the number of markers and the names of the  */
         /* individuals included in each analysis.                               */

         fprintf (stdout, "\n\n Markers = %d\n", NMarks);
         fprintf (oXtra, "Markers = %d\n", NMarks);

         fprintf (stdout, " Names:");
         fprintf (oXtra, " Individual names:");

         /* Each individual has 2 haplotypes, so for printing out names we will  */
         /* need an integer set to half of HapSetSize.                           */

         half_hss = HapSetSize / 2;

         for (hs_ndx = 0; hs_ndx < half_hss; ++hs_ndx)
         {
            fprintf (stdout, " %s", ObsNam[half_hss * k + hs_ndx]);
            fprintf (oXtra, " %s", ObsNam[half_hss * k + hs_ndx]);
         }

         fprintf (stdout, "\n");
         fprintf (oXtra, "\n");

         fflush (stdout);
         fflush (oXtra);

         /* If partially/un- phased data, collapse OUTPUT of 15 states down to 9.          */
         if (PhaseofData == PARTIALPHASE || PhaseofData == UNPHASED)
         {
            qq_ibd_final = alloc_dbl2d (NMarks, 9);
            for (i = 0; i < NMarks; ++i)
            {
               qq_ibd_final[i][0] = qq_ibd[i][0];
               qq_ibd_final[i][1] = qq_ibd[i][1];
               qq_ibd_final[i][2] = qq_ibd[i][2] + qq_ibd[i][3];
               qq_ibd_final[i][3] = qq_ibd[i][4];
               qq_ibd_final[i][4] = qq_ibd[i][5] + qq_ibd[i][6];
               qq_ibd_final[i][5] = qq_ibd[i][7];
               qq_ibd_final[i][6] = qq_ibd[i][8] + qq_ibd[i][9];
               qq_ibd_final[i][7] =
                  qq_ibd[i][10] + qq_ibd[i][11] + qq_ibd[i][12] + qq_ibd[i][13];
               qq_ibd_final[i][8] = qq_ibd[i][14];
            }
            nstmp = 9;
         }
         else
         {
            qq_ibd_final = qq_ibd;
            nstmp = ns;
         }

         if (!DebugGLDisp)
         {
            for (i = 0; i < NMarks; ++i)
            {
               fprintf (oXtra, " %*d  %10.6f ", nm, MarSel[i], snppos[i]);
               for (j = 0; j < nstmp; ++j)
                  fprintf (oXtra, " %7.4f", qq_ibd_final[i][j]);
               fprintf (oXtra, "\n");
            }
         }
         else
         {
            simStates = malloc (NMarks * sizeof (int));
            alloc_chk (WHERE, simStates);
            hap_sample (NMarks, ns, simStates, transmat, ffibd,
                        Mdist, pprob, PhaseofData);
            for (i = 0; i < NMarks; ++i)
            {
               fprintf (oXtra, " %d", simStates[i]);
            }
            fprintf (oXtra, "\n");
         }

      }
   }

   /* Close the extra output file.                                            */

   fflush (oXtra);
   fclose (oXtra);

   /* We are done, so all of the allocated storage can now be freed.          */

   free_1d ((void *) &duplicated_rows);
   free_2d ((void *) &state_matrix);

   free (snppos);
   free (haplodat);

   free (hapdat);
   free (allefreq);

   free (transmat);
   free (qq_ibd);
   free (pprob);

   if (PhaseofData == PARTIALPHASE)
   {
      free (phasdatindiv);
      free (phasdat);
      free (qq_ibd_final);
   }

   free_auto ();
   proc_free ();

   exit (EXIT_SUCCESS);
}
