/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

static double **prob_new = NULL;

void update_unlnkd_2_lnkd (int nummrks, int nped, double **prob, int ngenes,
                           int n_compnts, double p_lnkd)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
/* Probability of proposing a linked move:                                    */
     /* Metropolis-Hasting update: Proposing a move from an unlinked trait    */
     /*                            locus to a linked trait locus.             */
     /* 1. Propose a candidate marker interval, test position and cumulative  */
     /*    test position based upon the normalized weights:                   */
     /*    P(Y_T | S_{T-1}, S_{T+1}).                                         */
     /* NOTE - this was the old way when pseudo-priors were not being used.   */
     /*    However, with the use of pseudo-priors, this in effect gives       */
     /*    each test position an equal weighting.  This simplifies things     */
     /*    greatly, but comes at a computational cost.                        */
     /* 2. Generate S_T via the L sampler:                                    */
     /*    ie, P(Y_T | S_i, S_{i+1}, \lamba_T, \alpha)                        */
     /* 3. Calculate the M-H prob:                                            */
     /*    alpha ((S_T, \lamba, \alpha), (S_T*, \lamba*, \alpha*))            */
     /*           = p_unlinked * P(Y_T | S_i, S_j, \lamba*, \alpha*)          */
     /*                        * Prior(\lamba*, \alpha*))                     */
     /*  -------------------------------------------------------------------- */
     /* 1/(Num linked test positions) * P(Y_T | \lamba                        */
     /*             = 0.5, \alpha = 0) * Prior(0.5, 0)                        */
     /* 4. Randomly select move based on above probability.                   */
{
   int   c, ii;
   int   sx;
   int   ctposit;                       /* Candidate marker position within
                                           interval.                          */
   int   cintvl;                        /* Candidate marker interval.         */
   int   chrm_ctposit;                  /* Candidate marker position along
                                           the chromosome, required by Morgan. */
   int   cmlposit;                      /* Cumulative test position.          */

   double rnum;
   double prb_Yt_cndtl;                 /* log Pr(Y_T | S_i, S_i+1,
                                           \lamba, \alpha)                    */
   double m_h_prob;                     /* Metropolis-Hastings acceptance
                                           probability.                       */

   /* If not already allocated, allocate memory for the probabilities work    */
   /* matrix, prob_new, used in some of the subroutines which follow to       */
   /* update the position in, etc.                                            */

   if (prob_new == NULL)
      prob_new = alloc_dbl2d (nCompnt, NLoci);

   /* Each time the update_unlnkd_2_lnkd subroutine is called initialize the  */
   /* probabilities work matrix by copying the input parameter prob matrix    */
   /* into prob_new.                                                          */

   for (c = 0; c < n_compnts; c++)
      for (ii = 0; ii <= nummrks; ii++)
         prob_new[c][ii] = prob[c][ii];

   /*   +-----------------------------------------------------+               */
   /*   |   1. Propose candidate cumulative test position.    |               */
   /*   +-----------------------------------------------------+               */

   /* Randomly select linked test position [1, ..., NTPos].                   */
   chrm_ctposit = intran (1, NTPos);

   /* The following convoluted process derives the marker interval and        */
   /* test position within the marker interval in which the candidate         */
   /* CUMULATIVE test position resides.                                       */

   cmlposit = 0;
   for (ii = 0; ii <= nummrks; ii++)    /* Where there are nummrks + 1
                                           possible intervals.                */
   {
      if ((chrm_ctposit > cmlposit)
          && (chrm_ctposit <= cmlposit + NTPosInt[ii]))
      {
         cintvl = ii;
         ctposit = chrm_ctposit - cmlposit;
      }
      cmlposit += NTPosInt[ii];
   }

   /*   +------------------------------------------+                          */
   /*   |   2. Propose candidate S_T.              |                          */
   /*   +------------------------------------------+                          */

   for (sx = FEM; sx <= MAL; sx++)
      set_chromosomes (nummrks, Marek[sx], 0, &IntVal[chrm_ctposit],
                       &RecVal[sx][chrm_ctposit], sx);
   set_positions (nummrks);

   prb_Yt_cndtl = do_lsmpl (nummrks, 0, prob_new, ngenes);

   /*   +----------------------------------------------+                      */
   /*   |   3. Calculate M-H acceptance prob for move  |                      */
   /*   |         from unlinked to linked.             |                      */
   /*   +----------------------------------------------+                      */

   /*      alpha ((S_T, \lamba, \alpha),  (S_T*, \lamba*, \alpha*))           */
   /*            = p_unlinked  *  P(Y_T | S_i, S_j, \lamba*, \alpha*)         */
   /*              * Prior(\lamba*, \alpha*)                                  */
   /* ----------------------------------------------------------------------- */
   /* (1/Num linked test pos)  *  P(Y_T | \lamba = 0.5, \alpha = 0)           */
   /*                          *  Prior(0.5,0)                                */

   m_h_prob = (log (1.0 - p_lnkd) + prb_Yt_cndtl
               + log (PseudoPrior[chrm_ctposit]))
      - (log (1.0 / NTPos) + previous_realiz.prb_Yt_condtl
         + log (PseudoPrior[0]));

   m_h_prob = exp (m_h_prob);

   /*   +-----------------------------------------------+                     */
   /*   |   4. Randomly select move based on M-H prob.  |                     */
   /*   +-----------------------------------------------+                     */

   if (m_h_prob > 1)
      m_h_prob = 1;                     /* ie, min (1, M-H prob).             */
   rnum = rans ();

   if (rnum <= m_h_prob)                /* Accept move.                       */
   {
      /* Replace xx0 and xx1 in previous_realiz.                              */
      set_previous_realiz (cintvl, ctposit, RecVal[FEM][chrm_ctposit],
                           RecVal[MAL][chrm_ctposit], prb_Yt_cndtl, xx0,
                           xx1, nummrks, nped);

      /* Set prob.                                                            */
      for (ii = 0; ii <= nummrks; ii++)
         for (c = 0; c < n_compnts; c++)
            prob[c][ii] = prob_new[c][ii];

      /* Adjust tposit frequency.                                             */
      tposit_freq[chrm_ctposit]++;
   }

   else
   {
      /* Else move is rejected and previous values retained.                  */
      if (previous_realiz.intvl == 0)
         chrm_ctposit = 0;

      for (sx = FEM; sx <= MAL; sx++)
         set_chromosomes (nummrks, Marek[sx], 0, &IntVal[chrm_ctposit],
                          &RecVal[sx][chrm_ctposit], sx);
      set_positions (nummrks);

      /* Reassign xx0 and xx1.                                                */
      assign_xx (nummrks, nped);

      /* Reassign gg0 and gg1 based on reassigned xx0 and xx1.                */
      for (ii = 0; ii <= nummrks; ii++)
         set_FGL (ii);

      /* Adjust tposit frequency.                                             */
      tposit_freq[chrm_ctposit]++;
   }

   return;
}

static int *perm_sw = NULL;             /* Permutation arrays of switches     */
static int *perm_loc = NULL;            /* and loci.                          */

void update_loci (double prob_L, int nummrks, int nped, int nsw,
                  double **prob, int ng)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
/* Uses the L sampler or M sampler to update all loci and meioses,            */
/* respectively, via Gibbs proposals.                                         */
/*                                                                            */
/* nsw - number of meioses/switches.                                          */
/* ng  - number of genes at any given locus, based on pedigree structure.     */
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
{
   int   i, j, l, sx;
   int   c;                             /* Component index.                   */
   int   cloc;                          /* Locus being update.                */
   double rnum;                         /* Random number (0-1).               */
   double prb_Yt_cndtl;                 /* log Pr(Y_T | S_T-1, S_T+1)         */

   /* If not already allocated, allocate memory for the permutation arrays of */
   /* switches and loci.                                                      */
   if (perm_sw == NULL)
   {
      perm_sw = alloc_int1d (nsw);
      perm_loc = alloc_int1d (NLoci);
   }

   /* Each time update_loci is called, initialize the permutation arrays of   */
   /* switches and loci.                                                      */
   for (i = 0; i < nsw; ++i)
      perm_sw[i] = i;
   for (i = 0; i <= nummrks; ++i)
      perm_loc[i] = i;

   rnum = rans ();
   if (rnum <= prob_L)
   {
      /* +-------------------------+                                          */
      /* |     L SAMPLER           |                                          */
      /* +-------------------------+                                          */

      permute (perm_loc, nummrks + 1);
      for (j = 0; j <= nummrks; j++)
      {
         cloc = perm_loc[j];
         if (cloc != 0)                 /* Update the trait locus last.       */

            prb_Yt_cndtl = do_lsmpl (nummrks, cloc, prob, ng);
      }

      /* Update the trait locus and calculate
         log P(Y_T | S_i, S_i+1, \lamba, \alpha).                             */
      prb_Yt_cndtl = do_lsmpl (nummrks, 0, prob, ng);

      previous_realiz.prb_Yt_condtl = prb_Yt_cndtl; /* log P(Y_T | S_M)       */
   }

   else
   {
      /* +-------------------------+                                          */
      /* |     M SAMPLER           |                                          */
      /* +-------------------------+                                          */

      permute (perm_sw, nsw);
      for (j = 0; j < nsw; j++)
      {
         l = perm_sw[j] + 1;
         i = (meiosis[l].ind).index;
         sx = meiosis[l].sex;
         c = nghd[i].component - 1;

         do_msmpl (nummrks, 0, c, i, sx, prob, ng);
      }

      /* Calculate log Pr(Y_T | S_{T-1}, S_{T+1}) and why not update S_T at
         the same time.                                                       */
      prb_Yt_cndtl = do_lsmpl (nummrks, 0, prob, ng);

      previous_realiz.prb_Yt_condtl = prb_Yt_cndtl; /* log P(Y_T | S_M)       */
   }

   /* Store xx0 and xx1 in previous_realiz.                                   */
   for (i = 0; i <= nummrks; i++)
      for (j = 0; j < nped; j++)
      {
         previous_realiz.xx0[i][j] = xx0[i][j];
         previous_realiz.xx1[i][j] = xx1[i][j];
      }

   return;
}

void update_position (double p_lnkd, int window_size, int nummrks,
                      int nped, double **prob, int ngenes, int n_compnts)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
/*  M-H update for changing the position of the trait locus.                  */
/*  Three changes of position are possible:                                   */
/*    1. A change of trait locus position within a marker interval.           */
/*    2. A change of trait locus position along with a change in locus        */
/*       ordering (ie, the trait locus is moved to a new marker interval).    */
/*    3. Moving the trait locus to the unlinked case.                         */
/*  Note: window_size => total window width = 2 * window_size + 1.            */
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
{
   int   c, ii;
   int   sx;                            /* sex                                */
   int   cndt_posit;
   int   cndt_intvl;
   int   cuml_posit_new;
   int   cuml_posit_old;
   double q_old2new;
   double q_new2old;
   double prb_Yt_cndtl;                 /* log P(Y_T | S_{T-1}, S_{T+1})      */
   double M_H_prob;

   /* Copy the input parameter prob matrix into prob_new.                     */

   for (c = 0; c < n_compnts; c++)
      for (ii = 0; ii <= nummrks; ii++)
         prob_new[c][ii] = prob[c][ii];

   if (rans () <= p_lnkd)

      /* +-------------------------------------------+                        */
      /* |   MOVE TO LINKED POSITION PROPOSED        |                        */
      /* +-------------------------------------------+                        */

      /* Generate candidate position.                                         */

   {
      linked_proposal (window_size, previous_realiz.tposit,
                       previous_realiz.intvl, &cndt_posit,
                       &cndt_intvl, &cuml_posit_old, &cuml_posit_new,
                       &q_old2new, &q_new2old);

      /* Generate S_T, given candidate position.                              */
      for (sx = FEM; sx <= MAL; sx++)
         set_chromosomes (nummrks, Marek[sx], 0, &IntVal[cuml_posit_new],
                          &RecVal[sx][cuml_posit_new], sx);
      set_positions (nummrks);

      prb_Yt_cndtl = do_lsmpl (nummrks, 0, prob_new, ngenes);

      /* Calculate M-H ratio.                                                 */
      /* To avoid under/over flow errors, logs taken.                         */
      M_H_prob = (prb_Yt_cndtl + log (q_new2old)
                  + log (PseudoPrior[cuml_posit_new]))
         - (previous_realiz.prb_Yt_condtl + log (q_old2new)
            + log (PseudoPrior[cuml_posit_old]));

      M_H_prob = exp (M_H_prob);
   }

   else
      /* +-------------------------------------------+                        */
      /* |   MOVE TO UNLINKED POSITION PROPOSED      |                        */
      /* +-------------------------------------------+                        */

   {

      /* Generate candidate position.                                         */
      unlinked_proposal (&cndt_posit, &cndt_intvl, &cuml_posit_new);

      /* Generate S_T, given candidate position.                              */
      for (sx = FEM; sx <= MAL; sx++)
         set_chromosomes (nummrks, Marek[sx], 0, &IntVal[cuml_posit_new],
                          &RecVal[sx][cuml_posit_new], sx);
      set_positions (nummrks);

      prb_Yt_cndtl = do_lsmpl (nummrks, 0, prob_new, ngenes);

      /* Calculate M-H ratio for a move from linked to unlinked state.        */

      /* alpha ((S_T, \lamba, \alpha), (S_T*, \lamba*, \alpha*))
         = 1/(Num linked test pos) * P(Y_T | \lamba* = 0.5, \alpha* = 0)
         * Prior(0.5, 0).                                                     */
      /* -------------------------------------------------------------------- */
      /*  p_unlinked * P(Y_T | S_i, S_j, \lamba, \alpha)
       * Prior(\lamba, \alpha).                                               */

      /* Determine the current cumulative position.                           */
      cuml_posit_old = 0;
      for (ii = 0; ii < previous_realiz.intvl; ii++)
         cuml_posit_old += NTPosInt[ii];
      cuml_posit_old += previous_realiz.tposit;

      M_H_prob = (log (1.0 / NTPos) + prb_Yt_cndtl + log (PseudoPrior[0]))
         - (log (1.0 - p_lnkd) + previous_realiz.prb_Yt_condtl
            + log (PseudoPrior[cuml_posit_old]));

      M_H_prob = exp (M_H_prob);
   }

   /* Accept move with prob M_H_prob.                                         */

   if (M_H_prob > 1)
      M_H_prob = 1;
   if (rans () <= M_H_prob)
      /* Accept candidate move.                                               */
   {
      set_previous_realiz (cndt_intvl, cndt_posit, RecVal[FEM][cuml_posit_new],
                           RecVal[MAL][cuml_posit_new],
                           prb_Yt_cndtl, xx0, xx1, nummrks, nped);

      /* Set prob.                                                            */
      for (ii = 0; ii <= nummrks; ii++)
         for (c = 0; c < n_compnts; c++)
            prob[c][ii] = prob_new[c][ii];

      /* Adjust tposit frequency.                                             */
      tposit_freq[cuml_posit_new]++;
   }

   else
      /* Else move is rejected and previous values retained.                  */
   {

      /* NEED TO THINK ABOUT THIS /....... now that we have 0 IntVal
         not just implying unlinked case.                                     */

      if (previous_realiz.intvl == 0 && previous_realiz.rec_lft[FEM] == 0.5)
         cuml_posit_old = 0;            /* Unlinked.                          */
      else
      {
         cuml_posit_old = 0;
         for (ii = 0; ii < previous_realiz.intvl; ii++)
            cuml_posit_old += NTPosInt[ii];

         cuml_posit_old += previous_realiz.tposit;
      }
      for (sx = FEM; sx <= MAL; sx++)
         set_chromosomes (nummrks, Marek[sx], 0, &IntVal[cuml_posit_old],
                          &RecVal[sx][cuml_posit_old], sx);
      set_positions (nummrks);

      /* Reset xx0 and xx1 to previous.realiz.xx0 and previous.realiz.xx1.    */
      assign_xx (nummrks, nped);

      /* Based on reset xx0 and xx1, reset gg0 and gg1 to previous values.    */
      for (ii = 0; ii <= nummrks; ii++)
         set_FGL (ii);

      /* Adjust tposit frequency.                                             */
      tposit_freq[cuml_posit_old]++;
   }

   return;
}

void linked_proposal (int WindWdth, int posit_old, int intvl_old,
                      int *posit_new, int *intvl_new, int *cuml_posit_old,
                      int *cuml_posit_new, double *q_old2new, double *q_new2old)
{
   int   ii, dummy;

   int   X_upper;                       /* Upper limit of window.             */
   int   X_lower;                       /* Lower limit of window.             */
   int   h_upper;                       /* Upper width of window.             */
   int   h_lower;                       /* Lower width of window.             */

   /* Determine the old cumulative position - needed by Morgan routines.      */
   *cuml_posit_old = 0;
   for (ii = 0; ii < intvl_old; ii++)
      *cuml_posit_old += NTPosInt[ii];
   *cuml_posit_old += posit_old;

   /*   1. Determine if window extends past upper or lower test positions.    */
   /*   2. Sample a new canidate test position from the (adjusted) window.    */
   /*   3. Calculate q(old -> new).                                           */

   /* Test if window extends beyond right most test position.                 */
   if (*cuml_posit_old + WindWdth > NTPos) /* NTPos does not include
                                              unlinked pos (unlike NTPos1).   */
   {
      X_upper = NTPos;
      h_upper = NTPos - *cuml_posit_old;
   }
   else
   {
      X_upper = *cuml_posit_old + WindWdth;
      h_upper = WindWdth;
   }

   /* Test if window extends beyond the left most test position.              */
   if (*cuml_posit_old - WindWdth < 1)
   {
      X_lower = 1;
      h_lower = *cuml_posit_old - 1;
   }
   else
   {
      X_lower = *cuml_posit_old - WindWdth;
      h_lower = WindWdth;
   }

   /* Randomly accept position within interval.                               */
   *cuml_posit_new = intran (X_lower, X_upper);

   *q_old2new = 1.0 / (h_lower + h_upper + 1.0);

   /* Work out the new marker interval.                                       */
   *intvl_new = IntVal[*cuml_posit_new];

   dummy = *cuml_posit_new;
   for (ii = 0; ii < *intvl_new; ii++)  /* Derive new position within
                                           marker interval.                   */
      dummy -= NTPosInt[ii];
   *posit_new = dummy;

   /* 1. Given the new test position, determine if window extends             */
   /*    beyond test position boundaries.                                     */
   /* 2. Calculate  q(new -> old).                                            */

   /* Test if window extends beyond right most test position.                 */
   if ((*cuml_posit_new) + WindWdth > NTPos) /* NTPos does not include the
                                                unlinked position.            */
   {
      X_upper = NTPos;
      h_upper = NTPos - (*cuml_posit_new);
   }
   else
   {
      X_upper = (*cuml_posit_new) + WindWdth;
      h_upper = WindWdth;
   }

   /* Test if window extends beyond the left most test position.              */
   if ((*cuml_posit_new) - WindWdth < 1)
   {
      X_lower = 1;
      h_lower = (*cuml_posit_new) - 1;
   }
   else
   {
      X_lower = (*cuml_posit_new) - WindWdth;
      h_lower = WindWdth;
   }

   *q_new2old = 1.0 / (h_lower + h_upper + 1.0);

   return;
}

void unlinked_proposal (int *posit_new, int *intvl_new, int *cuml_posit_new)
{
   /* A deterministic move to the unlinked case.                              */
   *posit_new = 0;
   *intvl_new = 0;
   *cuml_posit_new = 0;

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_mh_proposals subroutine frees up any of the memory allocated by   */
/* one of the mh_proposals subroutines.                                       */
/*                                                                            */
/******************************************************************************/

void free_mh_proposals (void)
{
   free_2d ((void *) &prob_new);

   free_1d ((void *) &perm_sw);
   free_1d ((void *) &perm_loc);

   return;
}
