/******************************************************************************/
/*                                                                            */
/*  Program name: base_trait_lods.c (Calculates base trait lod scores using   */
/*                                   either pedigree peeling or Monte Carlo.  */
/*                                                                            */
/******************************************************************************/

#define PNAM  "base_trait_lods"
#define BASE_TRAIT_LODS

#include "lodscore.h"
#include "quant.h"

#include "base_trait_lods.h"
#include "base_trait_rtnes.c"

/******************************************************************************/
/*                                                                            */
/* The base_trait_lods main program.                                          */
/*                                                                            */
/******************************************************************************/

int main (int argc, char *argv[])
{
   FILE *oScorF = NULL;
   FILE *oSeedF = NULL;

   CTP_PTR p_ctp;

   /* Two local vectors needed to pass the lod score and its standard error   */
   /* to the prntLodScores subroutine.                                        */

   double pls_lod[2];                   /* Origin 1 variables needed to pass  */
   double pls_sd[2];                    /* results to prntLodScores.          */

   double overall_lod;                  /* Total lodscore over all components */
   double overall_sd;                   /* Total standard error over all      */
   /*                                      components.                        */

   double **prob;

   int   nCompnt1;                      /* Number of components to be         */
   /*                                      calculated exactly.                */
   int   comp_ndx;                      /* Component index from cmplist[] - 1. */
   int   comp_no;                       /* Current component number.          */

   int   nc;                            /* Character length of highest        */
   /*                                      numbered component.                */
   int   nn;                            /* Character length of the longest    */
   /*                                      first name in any component.       */
   int   np;                            /* Character length of most negative  */
   /*                                      log likelihood value.              */

   /* ntp is number of trait positions plus number of markers (NLocPos + 1).  */
   int   ntp;

   /* ................................  SET-UP   ............................ */

   PCleanUp = clean_base_trait;
   atexit (clean_morg);

   setupBasic ("LPR", argc, argv, PNAM);

   /* The restofAutozyg subroutine returns the number of switches.  I could   */
   /* save the number of switches here if there is any need for it.           */

   restofAutozyg ();
   setSampling ();

   nCompnt1 = 0;
   if (!GlobalMcmc)
   {
      /* Count the number of components to be calculated exactly.             */
      for (comp_no = 1; comp_no <= nCompnt; ++comp_no)
         if (cmplist[comp_no].nmeioses <= MaxNMexact)
            ++nCompnt1;
   }

   /* The call to setupInherit was moved down, so we can know whether exact   */
   /* computation is to be done for any of the components.                    */

   if (nCompnt1 == 0)
      prob = bt_setupInherit (1);
   else
      prob = bt_setupInherit (0);

   /* Now set evaluation points for lange_lik.                                */
   /* ntp is number of trait positions plus number of markers (NLocPos + 1).  */

   p_ctp = alloc_ctp (NLocPos + 1);

   ntp = set_eval_points (NMarks, p_ctp);

   if (ntp < NLocPos + 1)
      realloc_ctp (ntp, p_ctp);

   /* Open the output scores file to use as base_trait_lods' results file.    */

   if (oScorF == NULL)
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);

   /* Compute character lengths needed for the output to the scores file      */
   /* which follows.                                                          */

   nc = intWid (0, nCompnt);

   nn = 0;
   for (comp_ndx = 0; comp_ndx < nCompnt; ++comp_ndx)
      nn = max (strlen (pCompTab[comp_ndx].frstNam), nn);

   /* The character width wanted for np is the width of the most negative     */
   /* number generated for any log likelihood.  Here I have assumed that no   */
   /* log likelihood will be less than -999.0, so np will be 11.  Code needs  */
   /* to be added to make this code more general.                             */

   np = fldWid (-999.0, 0.0, 6);

   /* Call the alloc_btl_stor subroutine to allocate storage for the          */
   /* base_trait_lods_elem matrix.                                            */

   alloc_btl_stor ();

   if (NRealizations == 0)
   {
      /* If NRealizations is 0, then the user wants us to compute the base    */
      /* trait log-likelihood exactly using pedigree peeling.                 */

      /* Initialize trait segregations from pedigree data at trait locus.     */

      n_markers = NMarks;
      n_tloci = NTLocs;
      locus_peel_init (0, 1);
   }

   /* Call the set_quant_indices subroutine to initilize the quant_05 and the */
   /* quant_95 variables.  These will be used for subscripting into the rows  */
   /* of the btle_storage matrix to select the base_trait_lods_elments to     */
   /* output as the 5th percent and 95th percent quantiles.                   */

   set_quant_indices ();

   for (comp_no = 1; comp_no <= nCompnt; ++comp_no)
   {
      /* Determine the base log-likelihood for the current component in the   */
      /* pedigree.                                                            */

      printf ("\n\n");                  /* Two new lines before component.    */

      if (cmplist[comp_no].nmeioses == 0)
      {
         print_zero_meioses_message (comp_no);
         continue;
      }

      comp_ndx = cmplist[comp_no].index - 1;

      if (NRealizations > 0)
      {
         /* If Nrealizations is > 0, the base log-likelihood will be          */
         /* estimated using Monte Carlo.                                      */

         printf (" ====== Base trait LodScore for component %d", comp_ndx + 1);
         printf (" estimated using Monte Carlo ======\n\n");

         mc_lods (p_ctp, comp_no, comp_lod_scor, comp_sd);

         if (DebugLevel > 0)
         {
            /* Save the results returned from the mc_lods subroutine for the  */
            /* current component in local variables required for calling the  */
            /* prntLodScores subroutine.                                      */

            pls_lod[1] = comp_lod_scor[comp_ndx];
            pls_sd[1] = comp_sd[comp_ndx];

            prntLodSHead
               ("Base trait LodScore estimated using IID null realizations");

            prntLodScores (ntp, pls_lod, pls_sd);
         }
         fflush (stdout);

         /* If Nrealizations is > 0, we need to sort the current component's  */
         /* lod score contributions located in the base_trait_lods_elem       */
         /* matrix into ascending order (as defined by the comparBtls         */
         /* subroutine).                                                      */

         qsort (btle_storage[comp_ndx], NRealizations, sizeof (BTLE_ELEM),
                comparBtls);
      }
      else
      {
         /* If NRealizations is 0, then the user wants us to not use Monte    */
         /* Carlo but to only compute the base trait log-likelihood exactly   */
         /* using pedigree peeling.                                           */

         printf (" ====== Base trait log-likelihood for component %d",
                 comp_ndx + 1);
         printf (" computed using pedigree peeling ======\n\n");

         pp_lods (comp_no, pp_log_like, pp_lod_scor);

         if (DebugLevel > 0)
         {
            /* Save the result returned from the pp_lods subroutine for the    */
            /* current component in the vector allocated for this purpose.     */

            pls_lod[1] = pp_lod_scor[comp_ndx];

            prntLodSHead
               ("Base trait LodScore computed using pedigree peeling");
            printf ("                                      ");
            printf (" %11.4f\n", pls_lod[1]);
            printf ("\n\n");
         }
         fflush (stdout);

         comp_lod_scor[comp_ndx] = pp_lod_scor[comp_ndx];
      }

      /* Write the results to the user's terminal.                            */

      print_btl_comp_result (comp_no);

      /* Write the results to the output scores file.                         */

      if (oScorF != NULL)
      {
         fprintf (oScorF, " %*d", nc, comp_no);
         fprintf (oScorF, " %*s", nn, pCompTab[comp_ndx].frstNam);
         fprintf (oScorF, " %*f", np, comp_lod_scor[comp_ndx]);
         fprintf (oScorF, "\n");
         fflush (oScorF);
      }
   }

   /* Now print the overall cumulative base trait lod score unless there was  */
   /* just a single component in the pedigree.                                */

   if (nCompnt > 1)
   {
      /* Initialize the overall null lodscore and standard deviation to 0.0.  */

      overall_lod = 0.0;
      overall_sd = 0.0;

      for (comp_no = 1; comp_no <= nCompnt; ++comp_no)
      {
         /* Accumulate the overall null lodscore with its standard deviation. */

         comp_ndx = cmplist[comp_no].index - 1;
         overall_lod += comp_lod_scor[comp_ndx];
         overall_sd += comp_sd[comp_ndx];
      }

      printf ("\n\n ====== ");

      if (NRealizations > 0)
      {
         printf ("Base trait LodScore over all components");
         printf (" ======\n\n");

         printf ("                Monte Carlo    Monte Carlo\n");
         printf ("                lod score      StdErr\n");

         printf ("                %11.4f", overall_lod);
         printf ("    %11.4f\n", overall_sd);
      }
      else
      {
         printf ("Base trait log-likelihood over all components");
         printf (" ======\n\n");

         printf (" Pedigree peeling base 10 log-likelihood =");
         printf (" %11.4f\n", overall_lod);
      }

      printf ("\n\n");
      fflush (stdout);
   }

   /* Close the output scores file.                                           */

   fflush (oScorF);
   fclose (oScorF);

   /* The main portion of this program is now done.  So, if requested by the  */
   /* user, save the output sampler seed.                                     */

   if (NRealizations > 0 && SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   /* Free up allocated memory.                                               */

   free_btl_stor ();

   if (TraitTyp == QUANTITATIVE)
      free_input_data ();

   free_seg_pen (NLoci, &prob);
   free_chromo (0);

   free_ctp (p_ctp);

   /* Now we are done.                                                        */

   exit (EXIT_SUCCESS);
}
