/******************************************************************************/
/*                                                                            */
/* The bt_setupInherit subroutine calls functions to set up initial values    */
/* for inheritance indicators and Founder Gene Labels (FGL) prior to          */
/* preliminary realizations and the main Monte Carlo (MC) realizations.       */
/*                                                                            */
/* mm is 0 for sampling the trait, mm is 1 for markers only, no trait.        */
/*                                                                            */
/******************************************************************************/

double **bt_setupInherit (int mm)
{
   double **pr;

   int   i1;
   int   i2;

   /* Replace initInherit() with basic setup.                                 */
   /* Initialize maternal and paternal inheritance indicators
    * and founder genome labels (FGLs).
    * fncs:  set_FGL()
    * vars:  xx0[][], xx1[][], gg0[][], gg1[][]
    */

   for (i2 = 0; i2 < PedSize; ++i2)
      if (!(isNullNode (nghd[i2].pa)))
      {
         for (i1 = mm; i1 < NLoci; ++i1)
         {
            xx1[i1][i2] = 0;
            xx0[i1][i2] = 0;
         }
      }

   /* Set the initial FGLs:        gg0[][], gg1[][].                          */
   for (i1 = mm; i1 < NLoci; ++i1)
      set_FGL (i1);

   /* Set up peeling for founder genome labels.
    * fncs:  alloc_seg_pen(), alloc_gen_pen()
    * vars:  first[9*ngenes[j]], pp[2][ngenes[j]], obslist[2*NMarks][n], etc.
    * dirs:  Sample
    */
   /*  FGL peeling enabled for trait locus, even if not in MCMC: EAT 4/2010
    */
   setFGLpeel (0);

   /* Set up array for storing seg_pen results.
    * fncs:  seg_pen2()
    * vars:  prob[nCompnt][NMarks+1]
    * dirs:  Sample
    */

   pr = initPenProb (mm);

   return pr;
}

/******************************************************************************/
/*                                                                            */
/* The do_mclik_cmp subroutine obtains a single Monte Carlo estimate of the   */
/* trait log likelihood.  If the estimate is not a preliminary realization,   */
/* the contribution to the lod score is stored in the vector allocated for    */
/* saving the realizations in.                                                */
/*                                                                            */
/* The preliminary parameter is either TRUE or FALSE.  If it is TRUE, then    */
/* the realization computed by do_mclik_cmp is not to be used in the          */
/* estimation of the lod score or its standard error.                         */
/*                                                                            */
/* base_lik = base (ie, unlinked) likelihood.  This is computed once only.    */
/*                                                                            */
/******************************************************************************/

double do_mclik_cmp (CTP_PTR p_ctp, int cmp, int preliminary, double base_lik,
                     BTLE_PTR p_btle)
{
   double log_lik;                      /* The log likelihood returned by the */
   /*                                      gen_pen subroutine.                */
   double contribution;                 /* The contribution to the cumulative */
   /*                                      likelihood.                        */

   int  *genes[2];                      /* FGL's at the locus to the left of  */
   /*                                      the current test position.         */
   int   g_ndx;                         /* Subscript to the FGL's for the     */
   /*                                      current test position.             */
   int   err;

   /* Copy the appropriate FGL's to a local vector which is needed to pass to */
   /* the gen_pen subroutine.                                                 */

   g_ndx = p_ctp->loc_left[1] + max (NTLocs, 1) - 1;
   genes[0] = gg0[g_ndx];
   genes[1] = gg1[g_ndx];

   log_lik = gen_pen (cmp - 1, genes, loci + 0, nFGL, &err);

   if (preliminary)
      return log_lik;

   /* If the call is not a preliminary realization, compute the contribution  */
   /* of the realization to the estimate of the log-likehood and store it in  */
   /* the vector allocated for saving the realizations in.                    */

   contribution = exp (log_lik - base_lik);
   p_btle->contrib = contribution;

   return log_lik;
}

/******************************************************************************/
/*                                                                            */
/* Call the get_mc_realizations subroutine to obtain both the NPseudoPrIt and */
/* NRealizations IID realizations from the distribution of the base trait     */
/* log-likelihood.  The resulting preliminary log-likelihoods are stored in   */
/* the comp_log_like vector.  The resulting main log_likelihoods are stored   */
/* in the row of the base_trait_lods_elem matrix pointed to by p_btle.        */
/*                                                                            */
/******************************************************************************/

void get_mc_realizations (CTP_PTR p_ctp, int comp_no)
{
   BTLE_PTR p_btle;                     /* Working base_trait_lods_elem       */
   /*                                      pointer.                           */

   double base_lik;                     /* Current contribution to the base   */
   /*                                      log-likelihood.                    */

   int   comp_ndx;                      /* Current component subscript.       */
   int   realization;                   /* Current realization subscript.     */

   /* Calculate the subscript of the base_trait_lods_elem's matrix row to use */
   /* for the current component.                                              */

   comp_ndx = cmplist[comp_no].index - 1;

   /* ...............  PRELIMINARY REALIZATIONS  ............................ */

   /* Now do preliminary realizations.                                        */

   p_btle = btle_storage[comp_ndx];
   base_lik = 0.0;

   printf (" Now starting %d preliminary realizations\n", NPseudoPrIt);
   for (realization = 0; realization < NPseudoPrIt; ++realization)
   {
      realize_null (comp_no, 1, cmplist[comp_no].nmeioses); /* Marker 1 posn. */
      reset_FGL (1, comp_no);
      base_lik += do_mclik_cmp (p_ctp, comp_no, TRUE, base_lik, p_btle);
   }

   base_lik /= NPseudoPrIt;
   comp_log_like[comp_ndx] = base_lik;

   printf (" Crude log-likelihood from preliminary realizations = %e\n",
           base_lik);
   fflush (stdout);

   /* ...............  COMPUTE USING MONTE CARLO  ........................... */

   /* Now do the regular Monte Carlo sampling, and score the results.         */

   printf (" Beginning %d realizations of Monte Carlo\n", NRealizations);
   fflush (stdout);

   for (realization = 0; realization < NRealizations; ++realization)
   {
      /* Update xx1, xx0.                                                     */

      realize_null (comp_no, 1, cmplist[comp_no].nmeioses); /* Marker 1 posn. */
      reset_FGL (1, comp_no);

      if (DebugMain)
      {
         fprintf (stderr, "\nAt realization %d,\n", realization + 1);
         dbgXXGG ();
      }

      /* Compute the contribution to the Rao_Blackwellized score.             */

      p_btle = &btle_storage[comp_ndx][realization];
      do_mclik_cmp (p_ctp, comp_no, FALSE, base_lik, p_btle);
   }

   /* Print status that we have finished with the Monte Carlo.                */

   printf (" Monte Carlo realizations completed\n");
   printf ("\n");
   fflush (stdout);

   return;
}

/******************************************************************************/
/*                                                                            */
/* Call the mc_lods_results subroutine to compute a single Monte Carlo        */
/* estimate of the likelihood and its standard deviation.                     */
/*                                                                            */
/* Before the mc_lods_results subroutine is called, the get_mc_realizations   */
/* subroutine needs to have been called to compute the base trait log-        */
/* likelihood, base_lik.                                                      */
/*                                                                            */
/******************************************************************************/

void mc_lods_results (int comp_no, double base_lik, double *p_lod, double *p_sd)
{
   BTLE_PTR p_btle;                     /* Working base_trait_lods_elem       */
   /*                                      pointer.                           */

   double mc_prob;                      /* Cumulative null lodscore estimate. */
   double mc_sd;                        /* Cumulative Monte Carlo standard    */
   /*                                      deviation.                         */

   double contribution;                 /* The contribution to the cumulative */
   /*                                      likelihood.                        */
   double nrealizations;                /* Double precision NRealizations.    */
   double denom;                        /* Denominator in standard deviation. */

   int   comp_ndx;                      /* Current component subscript.       */
   int   realization;                   /* Current realization subscript.     */

   /* Initialize both the cumulative null lod score estimate, mc_prob, and    */
   /* the Monte Carlo standard deviation, mc_sd, to 0.0.                      */

   mc_prob = 0.0;
   mc_sd = 0.0;

   /* Calculate the subscript of the base_trait_lods_elem's matrix row to use */
   /* for the current component.                                              */

   comp_ndx = cmplist[comp_no].index - 1;

   /* Compute the total of the log-likelihood contributions and standard      */
   /* deviations.                                                             */

   for (realization = 0; realization < NRealizations; ++realization)
   {
      p_btle = &btle_storage[comp_ndx][realization];
      contribution = p_btle->contrib;
      mc_prob += contribution;
      mc_sd += (contribution * contribution);
   }

   /* Finish computing the Monte Carlo estimate of the likelihood and its     */
   /* standard deviation.  The NRealizations global variable is converted to  */
   /* double precision to prevent overflows while computing denom.            */

   nrealizations = (double) NRealizations;
   mc_prob = mc_prob / nrealizations;
   denom = nrealizations * nrealizations - nrealizations;
   mc_sd = sqrt ((mc_sd - nrealizations * mc_prob * mc_prob) / denom);

   /* Now compute the estimated null lodscore along with its Monte Carlo     */
   /* standard deviation.                                                    */

   p_lod[comp_ndx] = log10 (mc_prob) + base_lik / log (10.0);
   p_sd[comp_ndx] = mc_sd / (mc_prob * log (10.0));

   return;
}

/******************************************************************************/
/*                                                                            */
/* The mc_lods subroutine is a short subroutine called to set up for and call */
/* the get_mc_realizations and the mc_lods_results subroutines.               */
/*                                                                            */
/******************************************************************************/

void mc_lods (CTP_PTR p_ctp, int cmp, double *p_lod, double *p_sd)
{
   int   comp_ndx;                      /* Current component subscript.       */
   int   nctemp;                        /* Backup copy of original n_compnts. */

   nctemp = n_compnts;
   if (cmp)
      n_compnts = 1;

   /* Now call get_mc_realizations to compute Monte Carlo realizations.       */

   get_mc_realizations (p_ctp, cmp);

   /* Call the mc_lods_results subroutine to compute and return the mc_lods   */
   /* subroutine's final results.                                             */

   comp_ndx = cmplist[cmp].index - 1;
   mc_lods_results (cmp, comp_log_like[comp_ndx], p_lod, p_sd);

   n_compnts = nctemp;

   return;
}

/******************************************************************************/
/*                                                                            */
/* The pp_lods subroutine is called to compute the base trait log-likelihood  */
/* for a single component with component number passed as the first argument. */
/* This subroutine uses pedigree peeling, so the computations are exact.      */
/*                                                                            */
/* However, not all pedigrees can be successfully peeled.  See section 6.5    */
/* (page 90) of Elizabeth Thompson's book "Statistical Inference From Genetic */
/* Data on Pedigrees" for a discussion of when pedigree peeling becomes       */
/* infeasible.                                                                */
/*                                                                            */
/* Both the second and third arguments are pointers to vectors of double      */
/* precision floating point numbers.  Both of these vectors need to contain   */
/* nCompnt elements.  The base trait log-likelihood for the component indexed */
/* in the cmp'th cmplist entry returned by the locus_mjrpeel_cmp subroutine   */
/* is returned by pp_lods in the corresponding entry of the second argument.  */
/* This base trait log-likelihood converted to a base 10 logarithm will be    */
/* returned in the corresponding entry of the third argument.                 */
/*                                                                            */
/******************************************************************************/

void pp_lods (int cmp, double *p_base_lik, double *p_lod)
{
   double base_lik;                     /* Base log-likelihood returned by    */
   /*                                      the locus_mjrpeel_cmp subroutine.  */

   double lodscore;                     /* The component's lod score.         */

   int   comp_ndx;                      /* Current component subscript.       */
   int   nctemp;                        /* Backup copy of original n_compnts. */

   nctemp = n_compnts;
   if (cmp)
      n_compnts = 1;

   set_simple_trans (loci, -1, -1, 0.5, 0.5, 0.5, 0.5);

   if (TraitTyp == QUANTITATIVE)
      base_lik = locus_mjrpeel_cmp (cmp, loci[0].ve, 1.0, 0, loci);
   else
      base_lik = locus_mjrpeel_cmp (cmp, 1.0, 1.0, 0, loci);

   /* Checks for zero probability in peeling operation.                       */
   if (base_lik == -DBL_MAX)
      give_up_where (WHERE, ERS_zpp);

   comp_ndx = cmplist[cmp].index - 1;
   p_base_lik[comp_ndx] = base_lik;

   lodscore = base_lik / log (10.0);
   p_lod[comp_ndx] = lodscore;

   n_compnts = nctemp;

   return;
}

/******************************************************************************/
/*                                                                            */
/* The alloc_btl_stor subroutine is called to allocate storage for the        */
/* base_trait_lods_elem matrix and for the vectors needed to save cumulative  */
/* base log-likelihoods, lod scores and standard errors of the lod scores for */
/* each component.                                                            */
/*                                                                            */
/******************************************************************************/

void alloc_btl_stor (void)
{
   int   comp_no;                       /* Current component number.          */

   btle_storage = (BTLE_ELEM **) malloc (nCompnt * sizeof (BTLE_ELEM *));
   alloc_chk (WHERE, btle_storage);

   btle_storage[0] = (BTLE_ELEM *) malloc (nCompnt * NRealizations *
                                           sizeof (BTLE_ELEM));
   alloc_chk (WHERE, btle_storage[0]);

   for (comp_no = 1; comp_no < nCompnt; ++comp_no)
      btle_storage[comp_no] = btle_storage[comp_no - 1] + NRealizations;

   /* The alloc_dbl1d subroutine initializes the allocated storage to 0.0.    */

   comp_log_like = alloc_dbl1d (nCompnt);
   comp_lod_scor = alloc_dbl1d (nCompnt);
   comp_sd = alloc_dbl1d (nCompnt);

   /* The following two vectors will be used to save the results returned by  */
   /* the pp_lods subroutine.                                                 */

   pp_log_like = alloc_dbl1d (nCompnt);
   pp_lod_scor = alloc_dbl1d (nCompnt);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_btl_stor subroutine is called to free up the storage that was     */
/* allocated by the alloc_btl_stor subroutine.                                */
/*                                                                            */
/******************************************************************************/

void free_btl_stor (void)
{
   free_2d ((void *) &btle_storage);

   free_1d ((void *) &comp_log_like);
   free_1d ((void *) &comp_lod_scor);
   free_1d ((void *) &comp_sd);

   free_1d ((void *) &pp_log_like);
   free_1d ((void *) &pp_lod_scor);

   return;
}

/******************************************************************************/
/*                                                                            */
/* Call the set_quant_indices subroutine once to calculate the indices to use */
/* in the btle_storage matrix to find the 5% quantile and the 95% quantile.   */
/*                                                                            */
/* The quant_05 and quant_95 variables are declared in the base_trait_lods.h  */
/* header file.                                                               */
/*                                                                            */
/******************************************************************************/

void set_quant_indices (void)
{
   double interm_product;               /* Temporary storage for the product  */
   /*                                      of sample size and percentile.     */

   /* Calculate subscripts to the 5th percent and 95th percent quantiles.     */
   /* The subtraction of 1 is needed so that quant_x is the correct subscript */
   /* for an origin-0 vector.                                                 */

   interm_product = ((double) NRealizations) * 0.05;
   quant_05 = (int) (floor (interm_product + 0.5) - 1);

   interm_product = ((double) NRealizations) * 0.95;
   quant_95 = (int) (floor (interm_product + 0.5) - 1);

   return;
}

/******************************************************************************/
/*                                                                            */
/* Debugging subroutine called to print the lod score contributions in the    */
/* base_trait_lods_elem matrix.                                               */
/*                                                                            */
/******************************************************************************/

void print_btl_row (BTLE_PTR p_btle, double base_lik, int comp_no)
{
   double contribution;                 /* Current contribution to lod score. */

   int   per_line = 5;                  /* Number per line.                   */
   int   btle_ndx;                      /* base_trait_lods_elem subscript.    */

   printf (" Vector of lod score contributions for component %d,", comp_no);
   printf (" (with base log likelihood of %17.10f):", base_lik);

   for (btle_ndx = 0; btle_ndx < NRealizations; ++btle_ndx)
   {
      if (btle_ndx % per_line == 0)
         printf ("\n");

      contribution = p_btle->contrib;
      printf (" %19.10f", contribution);
      ++p_btle;
   }

   printf ("\n");

   return;
}

/******************************************************************************/
/*                                                                            */
/* The print_btl_comp_result subroutine is called by the main program once    */
/* for each component in order to print the main resulting numbers to the     */
/* user's terminal.                                                           */
/*                                                                            */
/******************************************************************************/

void print_btl_comp_result (int comp_no)
{
   BTLE_PTR p_btle;                     /* Working base_trait_lods_elem       */
   /*                                      pointer.                           */

   double base_lik;                     /* Current contribution to the base   */
   /*                                      log-likelihood.                    */
   double mc_prob;                      /* Null lod score contribution.       */
   double lod;                          /* Resulting lod score.               */

   int   comp_ndx;                      /* Component index from cmplist[] - 1 */

   comp_ndx = cmplist[comp_no].index - 1;
   base_lik = comp_log_like[comp_ndx];
   p_btle = btle_storage[comp_ndx];

   /* If debugging is being done, the print_btl_row subroutine can be called  */
   /* to print the contributions to the lod scores stored in the component's  */
   /* row of the base_trait_lods_elem matrix.                                 */

   if (DebugLevel == 10)
      print_btl_row (p_btle, base_lik, comp_no);

   if (NRealizations > 0)
   {
      printf ("\n");
      printf (" Number of      Monte Carlo    Monte Carlo     ");
      printf ("%s", "5th %       95th %\n");

      printf (" realizations   lod score      StdErr          ");
      printf ("quantile    quantile\n");

      printf ("\n");
      printf ("%11d", NRealizations);
      printf ("  %14.4f %14.4f", comp_lod_scor[comp_ndx], comp_sd[comp_ndx]);

      /* Determine and print the 5th percent quantile of the vector of Monte  */
      /* Carlo lod score realizations.                                        */

      p_btle = &btle_storage[comp_ndx][quant_05];
      mc_prob = p_btle->contrib;
      lod = log10 (mc_prob) + base_lik / log (10.0);
      printf ("  %11.4f", lod);

      /* Determine and print the 95th percent quantile of the vector of Monte */
      /* Carlo lod score realizations.                                        */

      p_btle = &btle_storage[comp_ndx][quant_95];
      mc_prob = p_btle->contrib;
      lod = log10 (mc_prob) + base_lik / log (10.0);
      printf (" %11.4f", lod);

      printf ("\n");
   }
   else
   {
      /* NRealizations is equal to 0.                                         */

      lod = pp_lod_scor[comp_ndx];

      printf ("\n Pedigree peeling base 10 log-likelihood = %11.4f\n", lod);
   }

   printf ("\n");
   fflush (stdout);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The comparBtls subroutine was written for use as a comparison subroutine   */
/* by the standard C library qsort and bsearch subroutines.  This subroutine  */
/* is called to specifically compare two base_trait_lods_elem lod scores.     */
/* These are declared in the base_trait_lods.h header file.                   */
/*                                                                            */
/******************************************************************************/

int comparBtls (const void *key, const void *elem)
{
   BTLE_PTR p_array_elem;
   BTLE_PTR p_key;

   int   result;

   p_array_elem = (BTLE_PTR const) elem;
   p_key = (BTLE_PTR const) key;

   /* Call the qfunc_doub subroutine to compare the lod score in the key with */
   /* the lod score in the array element.                                     */

   result = qfunc_doub (&p_key->contrib, &p_array_elem->contrib);

   return result;
}
