/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "nghds.h"

/*
 * Function to form the neighbourhoods for a pedigree with data matrix formed
 * by the input_data routine. The mother and father submatrices are formed
 * here and a sex check caried out. Existence of the nghd array is assumed
 * and these structures are all initialised to be null. The null node also
 * has to be initialised so that we have somewhere to send the unknown
 * parents of founders.
 * Routine by Charlie Geyer (1991) based on earlier routine of Nuala Sheehan.
 *
 * component indicator of individuals added; Sept 1993; EAT.
 */
static int **mother;                    /* was mother[MAX_SIZE][3] */
static int **father;                    /* was father[MAX_SIZE][3] */
static int *assigned;                   /* nghd'd or not  */

static int father_mat (int **, int **), mother_mat (int **, int **);
static void bounds (int, int **, int *), nullify (Ind *);
static void parent_sort (int **), make_nghd (int **);
static void node (int, int **, int *);

void nghds (void)
{
   int   i, bad_dads, bad_mums;
   KEYS_PP ppK;

   mother = alloc_int2d (pedsize, 3);
   father = alloc_int2d (pedsize, 3);

   assigned = malloc (pedsize * sizeof (int));
   alloc_chk (WHERE, assigned);
   for (i = 0; i < pedsize; ++i)
      assigned[i] = 0;

   /* Initialise all nodes. */
   nullify (&nghd[pedsize]);
   nullify (&nullnode);
   for (i = 0, ppK = ppKeyTab; i < pedsize; ++i, ++ppK)
   {
      strncpy (nghd[i].self, (*ppK)->nam, MAXNAMLEN + 1);
      nghd[i].index = data[i][0];
      nghd[i].sex = data[i][4];
      nghd[i].component = -1;

      if (data[i][2] == UNKNOWN && data[i][3] == UNKNOWN)
      {
         /* founders */
         nghd[i].pa = &nullnode;
         nghd[i].ma = &nullnode;
      }
      else if (data[i][2] == UNKNOWN || data[i][3] == UNKNOWN)
      {
         /* only has one parent! */
         give_up_where (WHERE, ERN_1pt, nghd[i].self);
      }
      else
      {
         nghd[i].pa = &nghd[data[i][2] - 1];
         nghd[i].ma = &nghd[data[i][3] - 1];
      }

      nghd[i].marriages = NULL;
      nghd[i].first_kid = &nullnode;
      nghd[i].next_patsib = &nullnode;
      nghd[i].next_matsib = &nullnode;
   }

   bad_dads = father_mat (data, father);
   parent_sort (father);

   bad_mums = mother_mat (data, mother);
   parent_sort (mother);

   if (bad_dads > 0 || bad_mums > 0)
   {
      if (bad_dads + bad_mums == 1)
         give_up_where (WHERE, ERN_sx1, bad_dads + bad_mums);
      else
         give_up_where (WHERE, ERN_sex, bad_dads + bad_mums);
   }

   /* Set marriages for those with children in the pedigree */
   make_nghd (father);
   make_nghd (mother);

   free_2d ((void *) &father);
   free_2d ((void *) &mother);

   free (assigned);

   /* now we can sort the key table by name for faster binary search to get an
    * index from a name */
   sortNam ();
}

/*
 * Function to make the neighbourhoods for all the individuals
 * in parent (i.e.the sorted mother/father matrix).
 * It calls on the functions bounds() and node().
 */
static void make_nghd (int **parent)
{
   int   k, limits[2], member_index;
   int   ktbIndex;

   /*
    * Run down through all the parents coded as 0 i. e. founders' parents.
    */
   for (k = 0; k < pedsize && parent[k][1] == UNKNOWN; k++)
      ;
   member_index = k;

   while (member_index < pedsize)
   {
      /* index in the Key Table, 0 .. pedsize - 1,
       * also the index in nghd */
      ktbIndex = parent[member_index][1] - 1;

      bounds (member_index, parent, limits);

      /*
       * Now have the two delimiting indices of the block of children
       * associated with the individual indexed by parent_index.
       */

      if (assigned[ktbIndex] == 0)
         /* Have no neighbourhood yet */
         node (ktbIndex, parent, limits);
      else
         give_warn (ERN_mng, "make_nghd", (*(ppKeyTab + ktbIndex))->nam);

      if (limits[1] < pedsize)
         member_index = limits[1] + 1;
   }
}

/*
 * Function to find the upper and lower indices of the block of children
 * associated with a given individual whose index in a, "number",
 * is already known.
 * The array a is the sorted mother or father submatrix formed by the
 * functions father_mat() and mother_mat() respectively.
 * These indices are put into a length-2 array, "limits".
 * The individual of interest is in the second column of the array, a,
 * and the spice in the third column are then sorted within the indices
 * returned.
 */
static void bounds (int number, int **a, int limits[2])
{
   int   i, j, k, temp1, temp2, temp3;

   for (i = number; i + 1 < pedsize && a[i + 1][1] == a[number][1]; i++)
      ;
   limits[1] = i;

   for (i = number; i - 1 >= 0 && a[i - 1][1] == a[number][1]; i--)
      ;
   limits[0] = i;

   /* sort spice within bounds found above. */

   for (j = limits[0]; j < limits[1]; j++)
      for (k = j + 1; k < limits[1] + 1; ++k)
         if (a[j][2] > a[k][2])
         {
            temp1 = a[j][0];
            temp2 = a[j][1];
            temp3 = a[j][2];
            a[j][0] = a[k][0];
            a[j][1] = a[k][1];
            a[j][2] = a[k][2];
            a[k][0] = temp1;
            a[k][1] = temp2;
            a[k][2] = temp3;
         }
}

/*
 * Function to find the neighbourhood of an individual in the pedigree who
 * is, himself, a parent.
 * The function takes three arguments: the name of the individual, the
 * sorted father or mother submatrix, "parent", (depending on whether
 * the individual is male or female), and the 2-dimensional array "limits"
 * returned by the "bounds" function giving the upper and lower indices
 * of the block of children in "parent" associated with the individual
 * in question.
 * The data matrix data[pedsize + 1][NcolData], sorted father or mother
 * submatrix parent[pedsize + 1][3], * and the array of structures of type
 * ind, nghd[pedsize + 1], are all available.  We also have the index of the
 * individual in parent (i.e.sorted father/ mother matrix), denoted by
 * parent_index.  It is assumed that the null node has been set up as a global
 * variable with its "self" component initialised to UNKNOWNNAME to cope with
 * calls to parents of founders.
 */
static void node (int row_index, int **parent, int limits[2])
{
   Indlist *child_ptr;
   Marlist *marriage_ptr;
   int   i, j;

   /* check for a valid index */
   if (row_index != INVALID_INDEX)
   {
      /*
       * Now have limits[0] and limits[1] s. t.
       * parent[limits[0]][0] is the row index in the data matrix for the
       * first child, and parent[limits[0]][2] is the index + 1 of the other
       * parent The matrix "parent" will be the sorted mother/father
       * submatrix.
       * We now set up the linked list of marriages and the linked list of
       * children within each marriage allocating space for the appropriate
       * structures when required.
       */
      i = limits[0];
      nghd[row_index].marriages = (Marlist *) malloc (sizeof (Marlist));
      alloc_chk (WHERE, nghd[row_index].marriages);
      nghd[row_index].marriages->current = (Mar *) malloc (sizeof (Mar));
      alloc_chk (WHERE, nghd[row_index].marriages->current);
      nghd[row_index].marriages->current->spouse = &nghd[parent[i][2] - 1];
      marriage_ptr = nghd[row_index].marriages;

      while (i < limits[1] + 1)
      {
         marriage_ptr->current->kids = (Indlist *) malloc (sizeof (Indlist));
         alloc_chk (WHERE, marriage_ptr->current->kids);
         marriage_ptr->current->kids->current = &nghd[parent[i][0]];
         child_ptr = marriage_ptr->current->kids;

         for (j = i + 1; j < limits[1] + 1 && parent[j][2] == parent[i][2]; ++j)
         {
            /* sibs share both parents */
            child_ptr->next = (Indlist *) malloc (sizeof (Indlist));
            alloc_chk (WHERE, child_ptr->next);
            child_ptr = child_ptr->next;
            child_ptr->current = &nghd[parent[j][0]];
         }
         child_ptr->next = NULL;

         i = j;                         /* Increment for the while loop */

         if (i != limits[1] + 1)
         {
            marriage_ptr->next = (Marlist *) malloc (sizeof (Marlist));
            alloc_chk (WHERE, marriage_ptr->next);
            marriage_ptr = marriage_ptr->next;
            marriage_ptr->current = (Mar *) malloc (sizeof (Mar));
            alloc_chk (WHERE, marriage_ptr->current);
            marriage_ptr->current->spouse = &nghd[parent[i][2] - 1];
         }
      }                                 /* end of while */

      marriage_ptr->next = NULL;
      /* Indicates that neighbourhood is made */
      assigned[row_index] = 1;
   }
   else
      /* Have an invalid row index - individual does not exist */
      give_up_where (WHERE, ERN_nop, "node", row_index);
}

/*
 * Function to bubble-sort the second column of a 3-column matrix in
 * ascending order(i.e.smallest in first row), and to apply the same
 * permutation to the first and third columns. This is used to sort the
 * mother submatrix on mothers and the father submatrix on fathers.
 */
static void parent_sort (int **a)
{
   int   i, j, temp1, temp2, temp3;

   for (i = 0; i < pedsize - 1; i++)
      for (j = i + 1; j < pedsize; j++)
         if (a[i][1] > a[j][1])
         {
            temp1 = a[i][0];
            temp2 = a[i][1];
            temp3 = a[i][2];
            a[i][0] = a[j][0];
            a[i][1] = a[j][1];
            a[i][2] = a[j][2];
            a[j][0] = temp1;
            a[j][1] = temp2;
            a[j][2] = temp3;
         }
}

/*
 * Function to make the nullnode for the neighbourhood routine.
 * Pointers to pa, ma and marriages are set to NULL.
 * It is assumed that a structure of type Ind has already been set up
 * nullnode and defined to be a global variable.
 */
static void nullify (Ind * nul)
{
   strncpy (nul->self, UNKNOWNNAME, MAXNAMLEN + 1);
   nul->index = INVALID_INDEX;
   nul->sex = UNSEXED;
   nul->pa = NULL;
   nul->ma = NULL;
   nul->marriages = NULL;
   nul->first_kid = NULL;
   nul->next_patsib = NULL;
   nul->next_matsib = NULL;
}

/*
 * Function to create the mother submatrix consisting of the index column,
 * the column of mother labels and the column of father labels.
 * A sex check is included at this point.
 * The function returns the number of mothers who have been given the
 * wrong sex.  (One case of mistaken sex could feature for as many times
 * as she has children.)
 * This matrix will then be sorted by mothers, the same permutation being
 * applied to the other two columns, which will enable us to pick off
 * the list of children associated with each mother.  (These are given by
 * the indices in the first column.)  The third column, the fathers, will
 * later be sorted by the bounds function for each mother.
 * The program will not die at this point if there is a mistake with gender.
 * It is of interest to find how many such mistakes there are.
 * The nghds function will die if this function returns a * positive number.
 */
static int mother_mat (int **matrix, int **mot)
{
   int   i;
   int   n = 0;
   KEYS_PP ppK;

   for (i = 0; i < pedsize; i++)
   {
      mot[i][0] = matrix[i][0];
      mot[i][1] = matrix[i][3];
      mot[i][2] = matrix[i][2];

      /* Checking gender is female. */

      if (mot[i][1] != UNKNOWN && nghd[mot[i][1] - 1].sex != FEMALE)
      {
         ppK = ppKeyTab + mot[i][1] - 1;
         give_warn_where (WHERE, ERN_sxm, "mother_mat", (*ppK)->nam);
         ++n;
      }
   }
   return n;
}

/*
 * Function to create the father submatrix consisting of the index column,
 * the column of father labels and the column of mother labels.
 * A sex check is included at this point.
 * The function returns the number of fathers who have been given the
 * wrong sex.  (One case of mistaken sex could feature for as many times
 * as he has children.)
 * This matrix will then be sorted by fathers, the same permutatiion being
 * applied to the other two columns, which will enable us to pick off
 * the list of children associated with each father.  (These are given by
 * the indices in the first column.)  The third column, the mothers, will
 * later be sorted by the bounds function for each father.
 * The program will not die at this point if there is a mistake with gender.
 * It is of interest to find how many such mistakes there are.
 * The nghds function will die if this function returns a * positive number.
 */
static int father_mat (int **matrix, int **fat)
{
   int   i;
   int   n = 0;
   KEYS_PP ppK;

   for (i = 0; i < pedsize; i++)
   {

      fat[i][0] = matrix[i][0];
      fat[i][1] = matrix[i][2];
      fat[i][2] = matrix[i][3];
      /* Checking gender is male. */
      if (fat[i][1] != UNKNOWN && nghd[fat[i][1] - 1].sex != MALE)
      {
         printf ("DEBUG: %d %d %d %d\n", i, fat[i][1],
                 nghd[fat[i][1] - 1].index, nghd[fat[i][1] - 1].sex);
         ppK = ppKeyTab + fat[i][1] - 1;
         give_warn_where (WHERE, ERN_sxf, "father_mat", (*ppK)->nam);
         ++n;
      }
   }
   return n;
}

/******************************************************************************/
/*                                                                            */
/* The partial_nghds subroutine is a small subset of the nghds subroutine.    */
/* It completes the initialization of the nghd structure sufficient enough    */
/* to allow programs to run using an input individuals file (which does not   */
/* contain information on the pedigree structure, but rather only names of    */
/* individuals, their component numbers and their trait data.                 */
/*                                                                            */
/* Existence of the nghd array is assumed and that these structures are all   */
/* initialized to be null.  The null node also has to be initialized.  Each   */
/* individual's component number is copied from the key table.                */
/*                                                                            */
/******************************************************************************/

void partial_nghds (void)
{
   KEYS_PP ppK;                         /* Key table pointer.                 */

   int   i1;

   /* Initialize all nodes.                                                   */
   nullify (&nghd[pedsize]);
   nullify (&nullnode);

   for (i1 = 0, ppK = ppKeyTab; i1 < pedsize; ++i1, ++ppK)
   {
      strncpy (nghd[i1].self, (*ppK)->nam, MAXNAMLEN + 1);
      nghd[i1].index = data[i1][0];
      nghd[i1].sex = data[i1][4];

      nghd[i1].pa = &nullnode;
      nghd[i1].ma = &nullnode;
      nghd[i1].marriages = NULL;
      nghd[i1].first_kid = &nullnode;
      nghd[i1].next_patsib = &nullnode;
      nghd[i1].next_matsib = &nullnode;

      /* Copy this individual's component number from the key table.          */
      nghd[i1].component = (*ppK)->comp;
   }

   /* Now we can sort the key table by name for faster binary search to get   */
   /* an index from a name.                                                   */
   sortNam ();

   return;
}
