/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 *   Setup functions for Morgan main programs:
 *     procParStat()
 *     setupBasic()
 */

#include "parsie.h"
#include "rans.h"
#include "quantset.h"

/**
 * Process the parameter file and set corresponding globals
 * (these start with capital letters and are found in globs.h).
 */

void procParStat (int argc, char **argv, char *ppn)
{

   /* parse input parameter statements  */
   if (!proc_pars (argc, argv, ppn))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* check statements for completeness and consistency.
    * Set up global parameters and data structures     */

   if (IS_MARKERDROP || IS_LM_AUTO || IS_LM_IBDTESTS || IS_LM_MAP || IS_LM_PVAL)
   {
      if (!proc_auto ())
         exit (EXIT_FAILURE);
   }

   if (IS_LM_BAYES || IS_LM_LODS || IS_LM_MARKERS || IS_LM_SCHNELL
       || IS_LM_TWOQTL)
   {
      if (!proc_lods ())
         exit (EXIT_FAILURE);
   }

#if 0
   /* check for 2-D graphical display option */
   if (IS_LM_AUTO)
      if (!proc_disp ())
         exit (EXIT_FAILURE);
#endif

   /* set up for PolyEM programs */
   if (IS_BIVAR || IS_MULTIVAR || IS_UNIBIG || IS_UNIVAR)
   {
      if (!proc_poly ())
         exit (EXIT_FAILURE);
   }

   /* set up pedigree file parameters */
   if (!proc_fils ())
      exit (EXIT_FAILURE);

   /* set up seed parameters */
   if (!proc_seeds ())
      exit (EXIT_FAILURE);

   prntStatusCP ();

   return;
}

/******************************************************************************/
/*                                                                            */
/* The init_rans_seeds subroutine initializes the appropriate seeds for the   */
/* rans() subroutine.                                                         */
/*                                                                            */
/******************************************************************************/

void init_rans_seeds (void)
{
   if (IS_LM_IBDTESTS && PermutEnable)
      setSeeds (PERM);
   else
   {
      /* If the program is not ibd_haplo set the sampler seeds.  If it is     */
      /* ibd_haplo then set the seeds only if realizations are being sampled. */

      if (!IS_IBD_HAPLO || NRealizations > 0)
         setSeeds (SAMPL);
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The first argument to this subroutine is a null-terminated string.  This   */
/* string consists of option code letters and is used to indicate which parts */
/* of the subroutine the caller wants executed.                               */
/*                                                                            */
/* If one of the option code letters is 'A', then the calling program wants   */
/* this subroutine to call proc_auto to perform semantic analysis.            */
/*                                                                            */
/* If one of the option code letters is 'L', then the calling program wants   */
/* this subroutine to call proc_lods to perform semantic analysis.            */
/*                                                                            */
/* If one of the option code letters is 'P', then the calling program will    */
/* require use of a pedigree.                                                 */
/*                                                                            */
/* If one of the option code letters is 'R', then the calling program will be */
/* using the rans() subroutine, which needs to have its seeds initialized.    */
/*                                                                            */
/* A 'D' option was added to setupBasic (to tell it to call proc_drop) while  */
/* working to develop the base_trait_lods program.  However I later decided   */
/* that proc_drop should not be called after all, so this option was never    */
/* completely debugged.  I still figured I might as well keep a 'D' option    */
/* here, just in case it might be needed later.                               */
/*                                                                            */
/* Also, a 'I' option was added to setupBasic (to tell it to call proc_ibd).  */
/* However, this option is not currently being used by any program and has    */
/* never been checked out.                                                    */
/*                                                                            */
/* Note that the calling program should only use at most one of the following */
/* option code letters: 'A', 'D', 'I' or 'L'.  It is okay to not include any  */
/* of these four option code letters.                                         */
/*                                                                            */
/******************************************************************************/

void setupBasic (char *options, int argc, char **argv, char *ppn)
{
   initMorgProg (argv);

   /* Parse input parameter statements.                                       */
   if (!proc_pars (argc, argv, ppn))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* Check statements for completeness and consistency and set up global     */
   /* parameters and data structures by calling proc_auto, proc_drop,         */
   /* proc_ibd or proc_lods, depending on which option code letters were      */
   /* passed by the program that has called setupBasic.                       */

   if (strchr (options, 'A'))
   {
      if (!proc_auto ())
         exit (EXIT_FAILURE);
   }
   else if (strchr (options, 'D'))
   {
      if (!proc_drop ())
         exit (EXIT_FAILURE);
   }
   else if (strchr (options, 'I'))
   {
      if (!proc_ibd ())
         exit (EXIT_FAILURE);
   }
   else if (strchr (options, 'L'))
   {
      if (!proc_lods ())
         exit (EXIT_FAILURE);
   }
   else
   {
      if (!test_auto_pars ())           /* Check computational parameters.    */
         if (proc_err ())
            exit (EXIT_FAILURE);
   }

   if (strchr (options, 'P'))
   {
      /* If the calling program requires use of a pedigree file, set up the   */
      /* pedigree file parameters and check the pedigree file for errors.     */

      if (!(strchr (options, 'A') || strchr (options, 'I') ||
            strchr (options, 'L')))
      {
         /* If none of proc_auto, proc_ibd or proc_lods has been called the   */
         /* mak_ped_dat subroutine will not have been called.  We still need  */
         /* to call the mak_ped_dat subroutine and handle any of its errors.  */

         if (!proc_input_ped ())
            exit (EXIT_FAILURE);
      }

      if (!proc_fils ())
         exit (EXIT_FAILURE);
   }

   if (strchr (options, 'R'))
   {
      /* Set up seed parameters.                                              */
      if (!proc_seeds ())
         exit (EXIT_FAILURE);
   }

   prntStatusCP ();

   if (strchr (options, 'P'))
   {
      /* Complete checking the pedigree file for errors and then set up the   */
      /* nghd array by calling setPed.                                        */

      chkPedFil ();

      if (!IS_BT_LODS)
         setPed (NULL);
   }
   else
   {
      /* Otherwise, we need to call proc_end so the program can be terminated */
      /* in case there were any errors requiring ending the program.          */

      if (!proc_end ())
         exit (EXIT_FAILURE);
   }

   if (strchr (options, 'R'))
   {
      /* Call the init_rans_seeds subroutine so that the seeds needed by the  */
      /* rans() function can be initialized.                                  */

      init_rans_seeds ();
   }

   return;
}
